using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Runtime.InteropServices;
using System.IO;
using System.Diagnostics;
using System.Reflection;
using System.Threading;

using MT.FPSource;

using DShowNET;
using DShowNET.Device;

namespace GraphController
{
	/// <summary>
	/// Zusammenfassung fr clsFilterGraphController.
	/// </summary>
	public class FilterGraphController
	{
		// --------------------- Events ----------------------

		/// <summary> Fired when a capture is completed (manually or automatically). </summary>
		public event EventHandler					CaptureComplete;



		// ------------- Protected/private Properties --------------

		protected GraphState		_graphState = GraphState.Null;		// State of the internal filter graph
		protected bool				_firstPreviewRendered = false;		// When _graphState==Rendered, have we rendered the preview stream after DeviceSelection?
		protected bool				_isPreviewRendered = false;			// When _graphState==Rendered, have we rendered the preview stream?
		protected bool				_isCaptureRendered = false;			// When _graphState==Rendered, have we rendered the capture stream?
		protected bool				_wantPreviewRendered = false;		// Do we need the preview stream rendered (VideoDevice and _previewWindow != null)
		protected bool				_wantCaptureRendered = false;		// Do we need the capture stream rendered
		protected bool				_wantCapture = false;				// Do we need to capture

		protected bool				_grabImageSaveState;

		protected int				_rotCookie = 0;						// Cookie into the Running Object Table
		protected Filter			_videoDevice = null;				// Property Backer: Video capture device filter
		protected Filter			_audioDevice = null;				// Property Backer: Audio capture device filter
		protected Filter			_videoCompressor = null;			// Property Backer: Video compression filter
		protected Filter			_audioCompressor = null;			// Property Backer: Audio compression filter
		protected string			_filename = "";						// Property Backer: Name of file to capture to
		protected Control			_previewWindow = null;				// Property Backer: Owner control for preview
		protected Image				_grabBitmap = null;					// Property Backer: Owner control for Grabbing
		protected VideoCapabilities	_videoCaps = null;					// Property Backer: capabilities of video device
		protected AudioCapabilities	_audioCaps = null;					// Property Backer: capabilities of audio device
		protected SourceCollection	_videoSources = null;				// Property Backer: list of physical video sources
		protected SourceCollection	_audioSources = null;				// Property Backer: list of physical audio sources
		protected PropertyPageCollection _propertyPages = null;			// Property Backer: list of property pages exposed by filters
		protected Tuner				_tuner = null;						// Property Backer: TV Tuner
		protected FPDetector		_grabVideo = null;
		protected SampleGrabber		_grabImage = null;
		protected IFilterGraph      _filterGraph;						// DShow Filter: filter Graph interface
		protected IBaseFilter       _nullRenderer;						// DShow Filter: Null Renderer
		protected IBaseFilter       _nullRenderer2;						// DShow Filter: Null Renderer
		protected IBaseFilter       _infTee;							// DShow Filter: Infinite Pin Tee Filter
		protected IBaseFilter		_transformFilter;					// DShow Filter: SWFilter Transform Filter
		protected IBaseFilter		_baseGrabFlt;						// DShow Filter: grabber filter interface.
		protected IBaseFilter       _sourceFilter;
		protected IMediaEventEx		_mediaEvt;							// DShow Filter: graph event interface
		protected ISampleGrabber	_sampGrabber;						// DShow Filter: Sample Grabber
		protected IGraphBuilder		_graphBuilder;						// DShow Filter: Graph builder 
		protected IMediaControl		_mediaCtrl;							// DShow Filter: Start/Stop the filter graph -> copy of _graphBuilder
		protected IVideoWindow		_videoWin;							// DShow Filter: Control preview window -> copy of _graphBuilder
		protected ICaptureGraphBuilder2		_capGraph = null;			// DShow Filter: building graphs for capturing video
		protected IAMStreamConfig	_videoStreamConfig = null;			// DShow Filter: configure frame rate, size
		protected IAMStreamConfig	_audioStreamConfig = null;			// DShow Filter: configure sample rate, sample size
		protected IBaseFilter		_devFilter = null;					// DShow Filter: selected video device
		protected IBaseFilter		_videoCompressorFilter = null;		// DShow Filter: selected video compressor
		protected IBaseFilter		_audioDeviceFilter = null;			// DShow Filter: selected audio device
		protected IBaseFilter		_audioCompressorFilter = null;		// DShow Filter: selected audio compressor
		protected IBaseFilter		_muxFilter = null;					// DShow Filter: multiplexor (combine video and audio streams)
		protected IFileSinkFilter	_fileWriterFilter = null;			// DShow Filter: file writer
		//protected VideoCaptureFrameEventArgs _lastFrame = null;
		protected int				_totalFramesCounted = 0;
		protected PictureBox		_pictureBox = null;
		protected ISWFilter			_swInterface;
		protected bool				_noneVideo = false;
		/// <summary> interface to get information and control video. </summary>
		protected IBasicVideo2		_basicVideo;
		/// <summary> audio interface used to control volume. </summary>
		protected IBasicAudio		_basicAudio;
		/// <summary> interface to single-step video. </summary>
		protected IVideoFrameStep	_videoStep;
		/// <summary> seek interface for positioning in stream. </summary>
		protected IMediaSeeking		_mediaSeek;
		/// <summary> seek interface to set position in stream. </summary>
		protected IMediaPosition	_mediaPos;

		protected IBaseFilter		_aviSplitter;


		public IPictureGrabber PictureGrabber
		{
			get{	return _grabImage;}
		}

		// ------------- Constructors/Destructors --------------

		/// <summary> 
		///  Create a new Capture object. 
		///  _videoDevice and _audioDevice can be null if you do not 
		///  wish to capture both audio and video. However at least
		///  one must be a valid device. Use the <see cref="Filters"/> 
		///  class to list available devices.
		///  </summary>


		protected FilterGraphController()
		{
		}

		public FilterGraphController(Filter videoDevice, Filter audioDevice)
		{			
			if ( videoDevice == null)
				Console.WriteLine( "The _videoDevice parameter must be set to a valid Filter.\n" );
			this._videoDevice = videoDevice;
			this._audioDevice = audioDevice;
			this._filename = getTempFilename();
			createGraph();

			_grabVideo = new FPDetector(_sampGrabber);
			_grabImage = new SampleGrabber(_sampGrabber);

			_grabImage.BeginCapture +=new GraphController.SampleGrabber.EmptyDelegate(_grabImage_BeginCapture);
			_grabImage.EndCapture += new GraphController.SampleGrabber.EmptyDelegate(_grabImage_EndCapture);

			//Deactivate SWFilter
			//Set/Get Parameters of the Interface
			SWFilterParams	parset;
			SWFilterParams	parget;

			//Deactivate Networkcommunication of SWFilter
			//and initialize Sync Value
			int hr = _swInterface.GetParams(out parset);
			if( hr < 0 )
				Marshal.ThrowExceptionForHR( hr );
					
			parset.intNet = 0;
			parset.intSync = -1;
			
			hr = _swInterface.SetParams(out parset);
			if( hr < 0 )
				Marshal.ThrowExceptionForHR( hr );

			hr = _swInterface.GetParams(out parget);
			if( hr < 0 )
				Marshal.ThrowExceptionForHR( hr );

			Console.WriteLine("IntSync Signal: " + parget.intSync.ToString());

			//Initiliaize Sync Value of FPDetector
			_grabVideo.Sync = -1;
		}

		public void SetSync(int Value)
		{

			try
			{
				//Set new Sync Value in FPDetector
				_grabVideo.Sync = Value;

				//Set/Get Parameters of the Interface
				SWFilterParams	parset;
				
				Console.WriteLine("{0} sets Sync on {1}",System.Threading.Thread.CurrentThread.Name, Value);
				//Deactivate Networkcommunication of SWFilter
				int hr = _swInterface.GetParams(out parset);
				if( hr < 0 )
					Marshal.ThrowExceptionForHR( hr );
				
				parset.intSync = Value;

				Console.WriteLine("SyncSignal wurde auf " + parset.intSync.ToString() + " erhht.");

				hr = _swInterface.SetParams(out parset);
				if( hr < 0 )
					Marshal.ThrowExceptionForHR( hr );
			}
			catch(Exception ex)
			{
				Console.Write(ex.Message);
			}
		}

		protected void _grabImage_BeginCapture()
		{
			_grabImageSaveState = _grabVideo.isEnabled;
			_grabVideo.Disable();
		}

		protected void _grabImage_EndCapture()
		{
			if (_grabImageSaveState)
				_grabVideo.Enable();
		}

		/// <summary> Destructor. Dispose of resources. </summary>
		~FilterGraphController()
		{
			Dispose();
		}

		/// <summary> Possible states of the interal filter graph </summary>
		protected enum GraphState
		{
			Null,			// No filter graph at all
			SetParamDevice, // During you change the Resolution or Framerate on the DeviceFilter
			Created,		// Filter graph created with device filters added
			Rendered,		// Filter complete built, ready to run (possibly previewing)
			Capturing		// Filter is capturing
		}



		// ------------------ Public Properties --------------------

		/// <summary> Is the class currently capturing. Read-only. </summary>
		public bool Capturing { get { return( _graphState==GraphState.Capturing ); } }

		/// <summary> Has the class been cued to begin capturing. Read-only. </summary>
		public bool Cued { get { return( _isCaptureRendered && _graphState==GraphState.Rendered ); } }

		/// <summary> Is the class currently stopped. Read-only. </summary>
		public bool Stopped { get { return( _graphState!=GraphState.Capturing ); } }

		public bool Capture
		{
			set
			{
				_wantCapture = value;
				if(value==false)
				{
					_graphState = GraphState.SetParamDevice;
				}
			}
			get
			{
				return _wantCapture;
			}
		}



		public string Filename 
		{ 
			get { return( _filename ); } 
			set 
			{ 
				assertStopped();
				if ( Cued )
					throw new InvalidOperationException( "The _filename cannot be changed once cued. Use Stop() before changing the _filename." );
				_filename = value; 
				if ( _fileWriterFilter != null )
				{
					string s;
					AMMediaType mt = new AMMediaType(); 
					int hr = _fileWriterFilter.GetCurFile( out s, mt );
					if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );
					if ( mt.formatSize > 0 )
						Marshal.FreeCoTaskMem( mt.formatPtr ); 
					hr = _fileWriterFilter.SetFileName( _filename, mt );
					if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );
				}
			} 
		}


		public void setFramerate(double framerate)
		{
			// Disable preview to avoid additional flashes (optional)
			Capture = true;

			bool GrabberEnabled = _grabVideo.isEnabled;
			_grabVideo.Disable();
			
			Control lPreviewWindow = PreviewWindow;
			PreviewWindow = null;

			FrameRate = framerate;

			// Restore previous preview setting
			Capture = false;
			PreviewWindow = lPreviewWindow;
			
			//Initialize _after_ graph has been rebuild
			_grabVideo.Initialize();
			_grabImage.Initialize();

			if(GrabberEnabled) 
				_grabVideo.Enable();

		}

		public void setResolution(Size Resolution)
		{
			// Disable preview to avoid additional flashes (optional)
			Capture = true;

			bool GrabberEnabled = _grabVideo.isEnabled;
			_grabVideo.Disable();
			
			
			Control lPreviewWindow = PreviewWindow;
			PreviewWindow = null;

			FrameSize = Resolution;
			
			// Restore previous preview setting			
			Capture = false;
			PreviewWindow = lPreviewWindow;

			//Initialize _after_ graph has been rebuild
			_grabVideo.Initialize();
			_grabImage.Initialize();

			if(GrabberEnabled) 
				_grabVideo.Enable();
		}


		///  To stop video preview, set this property to null. </para>
		/// </remarks>
		public virtual Control PreviewWindow
		{
			get { return( _previewWindow ); }
			set
			{
				assertStopped();
				derenderGraph();
				_previewWindow = value;
				_wantPreviewRendered = ( ( _previewWindow != null ) && ( _videoDevice != null ) );
				renderGraph();
				startPreviewIfNeeded();
			}
		}

		public Control PreviewWindowChange
		{
			get { return( _previewWindow ); }
			set
			{
				const int WS_CHILD			= 0x40000000;	
				const int WS_CLIPCHILDREN	= 0x02000000;
				const int WS_CLIPSIBLINGS	= 0x04000000;

				_previewWindow = value;
				if(_isPreviewRendered)
				{
					// Get the IVideoWindow interface
					_videoWin = (IVideoWindow) _graphBuilder;

					// Set the video window to be a child of the main window
					int hr = _videoWin.put_Owner( _previewWindow.Handle );
					if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );

					// Set video window style
					hr = _videoWin.put_WindowStyle( WS_CHILD | WS_CLIPCHILDREN | WS_CLIPSIBLINGS);
					if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );

					// Position video window in client rect of owner window
					_previewWindow.Resize += new EventHandler( onPreviewWindowResize );
					onPreviewWindowResize( this, null );

					// Make the video window visible, now that it is properly positioned
					hr = _videoWin.put_Visible( DsHlp.OATRUE );
					if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );
				}
			}
		}


		public void InitializeVideoGrabber()
		{
				_grabVideo.Initialize(); 
		}

		public void DisposeVideoGrabber()
		{
			if(_grabVideo != null)
			{
				_grabVideo.Disable();				
				_grabVideo = null;
			}
		}

		public void noneVideo()
		{
			if(_noneVideo == true)
			{
				_noneVideo = false;
			}
			else
			{
				_noneVideo = true;
			}
		}

		public Image grabVideo
		{
			get 
			{
				return(_grabBitmap);
			}
			
		}

		public PictureBox setPictureBox
		{
			set 
			{
				_pictureBox = value;
			}
		}


		/// <summary> if the videoPanel is resized, adjust video preview window. </summary>
		public void resizeVideoWindow()
		{
			if( _videoWin != null )
			{
				Rectangle rc = _previewWindow.ClientRectangle;
				_videoWin.SetWindowPosition( 0, 0, rc.Right, rc.Bottom );
			}
		}

		public FPDetector fpDetector
		{
			get { return _grabVideo; }
		}

		public ISWFilter swInterface
		{
			get { return _swInterface; }
		}

		public SortedList getResolutionCaps()
		{
			AMMediaType mediaType = null;
			VideoStreamConfigCaps caps = null;
			IntPtr pCaps = IntPtr.Zero;
			try
			{

				// Ensure this device reports capabilities
				int c, size;
				int hr = _videoStreamConfig.GetNumberOfCapabilities( out c, out size );
				if ( hr != 0 ) Marshal.ThrowExceptionForHR( hr );
				if ( c <= 0 ) 
					throw new NotSupportedException( "This video device does not report capabilities." );
				if ( size > Marshal.SizeOf( typeof( VideoStreamConfigCaps ) ) )
					throw new NotSupportedException( "Unable to retrieve video device capabilities. This video device requires a larger VideoStreamConfigCaps structure." );
				if ( c > 1 )
					Console.WriteLine("This video device supports " + c + " capability structures. Only the first structure will be used." );

				// Alloc memory for structure
				pCaps = Marshal.AllocCoTaskMem( Marshal.SizeOf( typeof( VideoStreamConfigCaps ) ) ); 

				// Retrieve first (and hopefully only) capabilities struct
				SortedList mySize = new SortedList();
				for (int i = 0; i < c; i++)
				{
					hr = _videoStreamConfig.GetStreamCaps( i, out mediaType, pCaps );
					if ( hr != 0 ) Marshal.ThrowExceptionForHR( hr );

					
					// Convert pointers to managed structures
					caps = (VideoStreamConfigCaps) Marshal.PtrToStructure( pCaps, typeof( VideoStreamConfigCaps ) );

					if (mediaType.formatType == FormatType.VideoInfo &&
						mediaType.subType == MediaSubType.RGB24)
					{
						int width = caps.MinOutputSize.Width;
						int height = caps.MinOutputSize.Height;
						mySize.Add(width,new Size(width, height));
					}

					
				}
				return mySize;
			}
			finally
			{
				if ( pCaps != IntPtr.Zero )
					Marshal.FreeCoTaskMem( pCaps ); pCaps = IntPtr.Zero;
				if ( mediaType != null )
					DsUtils.FreeAMMediaType( mediaType ); mediaType = null;
			}
		}

		public VideoCapabilities VideoCaps 
		{ 
			get 
			{ 
				if ( _videoCaps == null )
				{
					if ( _videoStreamConfig != null )
					{
						try 
						{ 
							_videoCaps = new VideoCapabilities( _videoStreamConfig ); 
						}
						catch ( Exception ex ) { Debug.WriteLine( "VideoCaps: unable to create _videoCaps." + ex.ToString() ); }
					}
				}
				return( _videoCaps ); 
			} 
		}

		/// <summary> 
		///  The video capture device filter. Read-only. To use a different 
		///  device, dispose of the current Capture instance and create a new 
		///  instance with the desired device. 
		/// </summary>
		public Filter VideoDevice { get { return( _videoDevice ); } }

		/// <summary> 
		///  The audio capture device filter. Read-only. To use a different 
		///  device, dispose of the current Capture instance and create a new 
		///  instance with the desired device. 
		/// </summary>
		public Filter AudioDevice { get { return( _audioDevice ); }	}


		public Filter VideoCompressor 
		{ 
			get { return( _videoCompressor ); } 
			set 
			{ 
				assertStopped();
//				if ( _videoCompressorFilter != null )
//					_graphBuilder.RemoveFilter( _videoCompressorFilter  );
//				destroyGraph();

				// Reload any property pages exposed by filters
				if ( _propertyPages != null ) _propertyPages.Dispose(); _propertyPages = null;

				_videoCompressor = value;
				_videoCompressorFilter = (IBaseFilter) Marshal.BindToMoniker( VideoCompressor.MonikerString );
				// Get the video compressor and add it to the filter graph
//				renderGraph();
//				startPreviewIfNeeded();
			}
		}
		

		/// <summary> 
		///  The audio compression filter. 
		/// </summary>
		/// <remarks>
		///  When this property is changed 
		///  the internal filter graph is rebuilt. This means that some properties
		///  will be reset. Set this property as early as possible to avoid losing 
		///  changes. This property cannot be changed while capturing.
		/// </remarks>
		public Filter AudioCompressor 
		{ 
			get { return( _audioCompressor ); } 
			set 
			{ 
				assertStopped();
				destroyGraph();
				_audioCompressor = value;
				renderGraph();
				startPreviewIfNeeded();
			}
		}

		/// <summary> 
		///  The current video source. Use Capture.VideoSources to 
		///  list available sources. Set to null to disable all 
		///  sources (mute).
		/// </summary>
		public Source VideoSource 
		{ 
			get { return( VideoSources.CurrentSource ); } 
			set { VideoSources.CurrentSource = value; } 
		}

		/// <summary> 
		///  The current audio source. Use Capture.AudioSources to 
		///  list available sources. Set to null to disable all 
		///  sources (mute).
		/// </summary>
		public Source AudioSource 
		{
			get { return( AudioSources.CurrentSource ); } 
			set { AudioSources.CurrentSource = value; }
		}



		public SourceCollection VideoSources
		{
			get 
			{ 
				if ( _videoSources == null )
				{
					try
					{
						if ( _videoDevice != null )
							_videoSources = new SourceCollection( _capGraph, _devFilter, true );
						else
							_videoSources = new SourceCollection();
					}
					catch ( Exception ex ) { Debug.WriteLine( "VideoSources: unable to create VideoSources." + ex.ToString() ); }
				}
				return ( _videoSources );
			}
		}
		
		public SourceCollection AudioSources
		{
			get 
			{ 
				if ( _audioSources == null )
				{
					try
					{
						if ( _audioDevice != null )
							_audioSources = new SourceCollection( _capGraph, _audioDeviceFilter, false );
						else
							_audioSources = new SourceCollection();
					}
					catch ( Exception ex ) { Debug.WriteLine( "AudioSources: unable to create AudioSources." + ex.ToString() ); }
				}
				return ( _audioSources );
			}
		}



		public PropertyPageCollection PropertyPages 
		{
			get
			{
				if ( _propertyPages == null )
				{
					try 
					{ 
						_propertyPages = new PropertyPageCollection( 
							_capGraph, 
							_devFilter, _audioDeviceFilter, 
							_videoCompressorFilter, _transformFilter, 
							VideoSources, AudioSources);
					}
					catch ( Exception ex ) { Debug.WriteLine( "PropertyPages: unable to get property pages." + ex.ToString() ); }

				}
				return( _propertyPages );
			}
		}

		public double FrameRate
		{
			get
			{
				long avgTimePerFrame = (long) getStreamConfigSetting( _videoStreamConfig, "AvgTimePerFrame");
				return( (double) 10000000 / avgTimePerFrame );
			}
			set
			{
				long avgTimePerFrame = (long) ( 10000000 / value );
				setStreamConfigSetting( _videoStreamConfig, "AvgTimePerFrame", avgTimePerFrame);
			}
		}

		public Size FrameSize
		{
			get
			{
				BitmapInfoHeader bmiHeader;
				bmiHeader = (BitmapInfoHeader) getStreamConfigSetting( _videoStreamConfig, "BmiHeader");
				Size size = new Size( bmiHeader.Width, bmiHeader.Height );
				return( size );
			}
			set
			{
				BitmapInfoHeader bmiHeader;
				bmiHeader = (BitmapInfoHeader) getStreamConfigSetting( _videoStreamConfig, "BmiHeader");
				bmiHeader.Width = value.Width;
				bmiHeader.Height = value.Height;
				setStreamConfigSetting( _videoStreamConfig, "BmiHeader", bmiHeader );
			}		
		}

		public void Cue()
		{
			assertStopped();

			// We want the capture stream rendered
			_wantCaptureRendered = true;

			// Re-render the graph (if necessary)
			renderGraph();

			// Pause the graph
			int hr = _mediaCtrl.Pause();
			if ( hr != 0 ) Marshal.ThrowExceptionForHR( hr ); 
		}

		/// <summary> Begin capturing. </summary>
		public void Start()
		{
			assertStopped();

			// We want the capture stream rendered
			_wantCaptureRendered = true;

			// Re-render the graph (if necessary)
			renderGraph();

			// Start the filter graph: begin capturing
			int hr = _mediaCtrl.Run();
			if ( hr != 0 ) Marshal.ThrowExceptionForHR( hr ); 

			// Update the state
			_graphState = GraphState.Capturing;
		}

		/// <summary> 
		///  Stop the current capture capture. If there is no
		///  current capture, this method will succeed.
		/// </summary>
		public void Stop()
		{
			_wantCaptureRendered = false;

			// Stop the graph if it is running
			// If we have a preview running we should only stop the
			// capture stream. However, if we have a preview stream
			// we need to re-render the graph anyways because we 
			// need to get rid of the capture stream. To re-render
			// we need to stop the entire graph
			if ( _mediaCtrl != null )
			{
				_mediaCtrl.Stop();
			}

			// Update the state
			if ( _graphState == GraphState.Capturing )
			{
				_graphState = GraphState.Rendered;
				if ( CaptureComplete != null )
					CaptureComplete( this, null );
			}

			// So we destroy the capture stream IF 
			// we need a preview stream. If we don't
			// this will leave the graph as it is.
			try { renderGraph(); } 
			catch {}
			try { startPreviewIfNeeded(); } 
			catch {}
		}

		/// <summary> 
		///  Calls Stop, releases all references. If a capture is in progress
		///  it will be stopped, but the CaptureComplete event will NOT fire.
		/// </summary>
		public void Dispose()
		{
			_wantPreviewRendered = false;
			_wantCaptureRendered = false;
	
			_grabVideo = null;
			_pictureBox = null;
			_grabImage = null;
			CaptureComplete = null;
			_wantCapture = true;


			try { destroyGraph(); } 
			catch {}

			if ( _videoSources != null )
				_videoSources.Dispose(); _videoSources = null;
			if ( _audioSources != null )
				_audioSources.Dispose(); _audioSources = null;

		}
		public IAMVideoProcAmp IAMVideoProcAmpInit()
		{
			
			int hr;
			object comObj = null;
			ICaptureGraphBuilder2 bld = _capGraph;
			IBaseFilter flt = _devFilter;


			try 
			{
				Guid cat = PinCategory.Preview;
				Guid type = MediaType.Video;
				Guid iid = typeof(IAMVideoProcAmp).GUID;
				hr = bld.FindInterface(ref cat, ref type, flt, ref iid, out comObj);
				if( hr != 0 )
				{
					return null;
				}

				IAMVideoProcAmp videoProc=null;
				videoProc=comObj as IAMVideoProcAmp;
				if(videoProc == null) 
				{
					Console.WriteLine("IAMVideoProcAmp not present");
					return null;
				}
				
				return videoProc;

			}
			catch( Exception ee )
			{
				Console.WriteLine( "!Ds.NET: ShowPrevPinDialog " + ee.Message );
				return null;
			}
			finally
			{
				if( comObj != null )
					Marshal.ReleaseComObject( comObj ); comObj = null;
			}
		}

		public IAMVideoControl IAMVideoControlInit()
		{
			
			int hr;
			object comObj = null;
			ICaptureGraphBuilder2 bld = _capGraph;
			IBaseFilter flt = _devFilter;


			try 
			{
				Guid cat = PinCategory.Preview;
				Guid type = MediaType.Video;
				Guid iid = typeof(IAMVideoControl).GUID;
				hr = bld.FindInterface(ref cat, ref type, flt, ref iid, out comObj);
				if( hr != 0 )
				{
					return null;
				}

				IAMVideoControl videoControl=null;
				videoControl=comObj as IAMVideoControl;
				if(videoControl == null) 
				{
					Console.WriteLine("IAMVideoControl not present");
					return null;
				}
				
				return videoControl;

			}
			catch( Exception ee )
			{
				Console.WriteLine( "!Ds.NET: ShowPrevPinDialog " + ee.Message );
				return null;
			}
			finally
			{
				if( comObj != null )
					Marshal.ReleaseComObject( comObj ); comObj = null;
			}
		}

		public IAMStreamConfig IAMStreamConfigInit()
		{
			int hr;
			object comObj = null;
			ICaptureGraphBuilder2 bld = _capGraph;
			IBaseFilter flt = _devFilter;

			try 
			{

				Guid cat  = PinCategory.Capture;
				Guid type = MediaType.Interleaved;
				Guid iid = typeof(IAMStreamConfig).GUID;

				hr = bld.FindInterface( ref cat, ref type, flt, ref iid, out comObj );
				if( hr != 0 )
				{
					type = MediaType.Video;
					hr = bld.FindInterface( ref cat, ref type, flt, ref iid, out comObj );
					
					if( hr != 0 )
						return null;
				}
				
		
				//Instead of calling OleCreatePropertyFrame do ...
				IAMStreamConfig streamConfig=null;
				streamConfig=comObj as IAMStreamConfig;

				if(streamConfig == null) 
				{
					Console.WriteLine("IAMStreamConfig not present");
					return null;
				}

				return streamConfig;

			}
			catch( Exception ee )
			{
				Console.WriteLine( "!Ds.NET: IAMStreamConfig " + ee.Message );
				return null;
			}
			finally
			{
				if( comObj != null )
					Marshal.ReleaseComObject( comObj); comObj = null;
			}
		}

		public void VideoControl(IAMVideoControl videoControl)
		{
			IPin pinOut;
			Size test = new Size(160,120);
			int listsize;
			long[] framerates = new long[6];

			int hr= DsUtils.GetPin(_devFilter,PinDirection.Output,0, out pinOut);
			if( hr < 0 )
				Marshal.ThrowExceptionForHR(hr);

			VideoControlFlags flagcontrol;
			videoControl.GetCaps(pinOut,out flagcontrol);
			videoControl.GetMode(pinOut,out flagcontrol);
			Console.WriteLine(flagcontrol.ToString());
			long frame;
			//videoControl.GetCurrentActualFrameRate(pinOut,out frame);
			videoControl.GetMaxAvailableFrameRate(pinOut,0, test, out frame);
			videoControl.GetFrameRateList(pinOut,0,test,out listsize, framerates);
			videoControl.SetMode(pinOut, VideoControlFlags.Trigger);

			videoControl.GetMode(pinOut,out flagcontrol);
			Console.WriteLine(flagcontrol.ToString());
			Console.WriteLine("Max Available Framerate: " + frame.ToString());
			for(int i=0; i<listsize;i++)
			{
				Console.WriteLine("Framerate " + i + ": " + framerates[i].ToString());
			}
		}

		/// <summary> graph event (WM_GRAPHNOTIFY) handler. </summary>
		public virtual void OnGraphNotify()
		{
			if( _mediaEvt != null )
			{
				DsEvCode	code;
				int p1, p2, hr = 0;
				do
				{
					hr = _mediaEvt.GetEvent( out code, out p1, out p2, 0 );
					if( hr < 0 )
						break;
					hr = _mediaEvt.FreeEventParams( code, p1, p2 );
				}
				while( hr == 0 );
			}
			
		}

		public IPin getDevicePin()
		{
			IPin pinOut;
			int hr= DsUtils.GetPin(_devFilter,PinDirection.Output,0, out pinOut);
			if( hr < 0 )
				Marshal.ThrowExceptionForHR(hr);
			return pinOut;
		}

		public string SaveCaptureFile()
		{
			SaveFileDialog sd = new SaveFileDialog();
			sd.FileName = @"Capture.avi";
			sd.Title = "Save Video Stream as...";
			sd.Filter = "Video file (*.avi)|*.avi";
			sd.FilterIndex = 1;
			if( sd.ShowDialog() != DialogResult.OK )
			{ 
				return @"c:\capture.avi"; 
			}
			return sd.FileName;
		}


		// --------------------- Private Methods -----------------------
		
		/// <summary> 
		///  Create a new filter graph and add filters (devices, compressors, 
		///  misc), but leave the filters unconnected. Call renderGraph()
		///  to connect the filters.
		/// </summary>
		/// 
		/// <summary> create the used COM components and get the interfaces. </summary>
		protected virtual bool GetInterfaces()
		{
			Type comtype = null;
			object comobj = null;
			try 
			{

				// Make a new filter graph
				comtype = Type.GetTypeFromCLSID( Clsid.FilterGraph );
				if( comtype == null )
					throw new NotImplementedException( @"DirectShow FilterGraph not installed/registered!" );
				comobj = Activator.CreateInstance( comtype );
				_graphBuilder = (IGraphBuilder) comobj; comobj = null;

				// Get the Capture Graph Builder
				Guid clsid = Clsid.CaptureGraphBuilder2;
				Guid riid = typeof(ICaptureGraphBuilder2).GUID;
				comobj = DsBugWO.CreateDsInstance( ref clsid, ref riid );
				Console.WriteLine("Capture Graph wurde erzeugt");
				_capGraph = (ICaptureGraphBuilder2) comobj; comobj = null;

				// Get the AVISplitter
				comtype = Type.GetTypeFromCLSID( Clsid.AviSplitter);
				if( comtype == null )
					throw new NotImplementedException( @"DirectShow SampleGrabber not installed/registered!" );
				comobj = Activator.CreateInstance( comtype );
				Console.WriteLine("AVI Splitter wurde erzeugt");
				_aviSplitter = (IBaseFilter) comobj; comobj = null;

				// Get the Sample Grabber
				comtype = Type.GetTypeFromCLSID( Clsid.SampleGrabber );
				if( comtype == null )
					throw new NotImplementedException( @"DirectShow SampleGrabber not installed/registered!" );
				comobj = Activator.CreateInstance( comtype );
				Console.WriteLine("SampleGrabber wurde erzeugt");
				_sampGrabber = (ISampleGrabber) comobj; comobj = null;

				// Get the Infinite Tee Pin Filter
				comtype = Type.GetTypeFromCLSID( Clsid.InfTee);
				if( comtype == null )
					throw new NotSupportedException( "DirectX (8.1 or higher) not installed?" );
				comobj = Activator.CreateInstance(comtype);
				Console.WriteLine("Infinite Pin Tee Filter wurde erzeugt");
				_infTee = (IBaseFilter) comobj; comobj = null;

				// Get a Null Renderer
				comtype = Type.GetTypeFromCLSID( Clsid.NullRenderer);
				if( comtype == null )
					throw new NotSupportedException( "DirectX (8.1 or higher) not installed?" );
				comobj = Activator.CreateInstance(comtype);
				Console.WriteLine("Null Renderer wurde erzeugt");
				_nullRenderer = (IBaseFilter) comobj; comobj = null;

				// Get another Null Renderer
				comtype = Type.GetTypeFromCLSID( Clsid.NullRenderer);
				if( comtype == null )
					throw new NotSupportedException( "DirectX (8.1 or higher) not installed?" );
				comobj = Activator.CreateInstance(comtype);
				Console.WriteLine("Null Renderer wurde erzeugt");
				_nullRenderer2 = (IBaseFilter) comobj; comobj = null;

				// Get the Transform Tracking Filter
				comtype = Type.GetTypeFromCLSID( Clsid.SWFilter );
				if( comtype == null )
					throw new NotSupportedException( "DirectX (8.1 or higher) not installed?" );
				comobj = Activator.CreateInstance(comtype);
				Console.WriteLine("Transform Filter wurde erzeugt");
				_transformFilter = (IBaseFilter) comobj; comobj = null;

				//Interfaces to Control a Video Clip or Stream
				_mediaCtrl		= (IMediaControl)	_graphBuilder;
				_videoWin		= (IVideoWindow)	_graphBuilder;
				_mediaEvt		= (IMediaEventEx)	_graphBuilder;

				//Special Interfaces
				_baseGrabFlt	= (IBaseFilter)		_sampGrabber;
				_filterGraph	= (IFilterGraph)	_graphBuilder;
				_swInterface	= (ISWFilter)		_transformFilter;

				//Interfaces special to play a VideoClip
				_mediaSeek	    = (IMediaSeeking)	_graphBuilder;
				_mediaPos	    = (IMediaPosition)	_graphBuilder;
				_basicVideo	    = (IBasicVideo2)	_graphBuilder;
				_basicAudio	    = (IBasicAudio)		_graphBuilder;

				//*************************************************
				// Grab Init
				//*************************************************
				AMMediaType media = new AMMediaType();
				media.majorType	= MediaType.Video;
				media.subType	= MediaSubType.RGB24;
				media.formatType = FormatType.VideoInfo;		// ???
				int hr = _sampGrabber.SetMediaType( media );
				if( hr < 0 )
					Marshal.ThrowExceptionForHR( hr );
				//*************************************************
				// END Grab Init
				//*************************************************


				return true;
			}
			catch( Exception ee )
			{
				Console.WriteLine("Could not get interfaces\r\n" + ee.Message );
				return false;
			}
			finally
			{
				if( comobj != null )
					Marshal.ReleaseComObject( comobj ); comobj = null;
			}
		}

		/// <summary> build the capture graph. </summary>
		protected virtual bool SetupGraph(GraphState graphState)
		{
			int hr;
			int nextPin;
			IPin					pinOut;
			IPin					pinIn;
			IPin					pinOutRender;

			try 
			{
				if((int)graphState < (int)GraphState.SetParamDevice)
				{
					hr = _capGraph.SetFiltergraph( _graphBuilder );
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );
					
					hr = _graphBuilder.AddFilter( _devFilter, "Ds.NET Video Capture Device" );
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );

					AMMediaType media = new AMMediaType();
					_videoStreamConfig.GetFormat(out media);
					VideoInfoHeader videoInfoHeader = (VideoInfoHeader) Marshal.PtrToStructure( media.formatPtr, typeof(VideoInfoHeader) );
					videoInfoHeader.BmiHeader.Width=640; 
					videoInfoHeader.BmiHeader.Height=480;
					Marshal.StructureToPtr(videoInfoHeader, media.formatPtr, true);
					_videoStreamConfig.SetFormat(media); 
				}


				if(graphState != GraphState.Capturing)
				{

					hr= DsUtils.GetPin(_devFilter,PinDirection.Output  ,0, out pinOut);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);

					hr = _graphBuilder.AddFilter( _infTee, "Ds.NET Infinite Pin Tee Filter" );
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );
				
					hr= DsUtils.GetPin(_infTee,PinDirection.Input ,0, out pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);

					hr = _graphBuilder.Connect(pinOut, pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );

					//***********************************************************
					//  SampleGrabber Filter
					//***********************************************************
					 hr = DsUtils.NextPin(_infTee,PinDirection.Output,out nextPin);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);

					hr= DsUtils.GetPin(_infTee,PinDirection.Output,nextPin-1, out pinOut);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);
				
					hr = _graphBuilder.AddFilter(_nullRenderer,"Null Renderer");
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );

					//Add SampleGrabber Filter 
					hr = _graphBuilder.AddFilter(_baseGrabFlt, "Ds.NET Grabber");
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );			
					//Get Input Pin from SampleGrabber Filter
					hr= DsUtils.GetPin(_baseGrabFlt, PinDirection.Input,0, out pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);	
			
					hr = _graphBuilder.Connect(pinOut, pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );	
				
					hr= DsUtils.GetPin(_baseGrabFlt, PinDirection.Output,0, out pinOutRender);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);
				
					hr= DsUtils.GetPin(_nullRenderer, PinDirection.Input ,0,out pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);

					hr = _graphBuilder.Connect(pinOutRender, pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );
					//*************************************************
					// END SampleGrabber Filter
					//*************************************************

					//***********************************************************
					//  Transform Filter
					//***********************************************************
					hr = DsUtils.NextPin(_infTee,PinDirection.Output,out nextPin);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);

					hr= DsUtils.GetPin(_infTee,PinDirection.Output,nextPin-1, out pinOut);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);

					hr = _graphBuilder.AddFilter(_nullRenderer2,"Null Renderer");
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );

					//Add Transform Filter
					hr = _graphBuilder.AddFilter(_transformFilter,"SWFilter");
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );

					//Get Input Pin from Transform Filter
					hr= DsUtils.GetPin(_transformFilter, PinDirection.Input,0, out pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);		
								
					hr = _graphBuilder.Connect(pinOut, pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );

					hr= DsUtils.GetPin(_transformFilter, PinDirection.Output,0,out pinOut);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);

					hr= DsUtils.GetPin(_nullRenderer2, PinDirection.Input ,0,out pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);

					hr = _graphBuilder.Connect(pinOut, pinIn);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );
				}
				//*************************************************
				// END Transform Filter
				//*************************************************
				Console.WriteLine("Aktueller Aufbau des Filter Graphs:");
				hr = DsUtils.EnumFilters(_filterGraph);
				if( hr < 0 )
					Marshal.ThrowExceptionForHR( hr );


				return true;
			}
			catch( Exception ee )
			{
				Console.WriteLine("Could not setup graph\r\n" + ee.Message );
				return false;
			}
			finally
			{

			}
		}

		protected virtual void createGraph()
		{
			Guid					cat;
			Guid					med;
			int						hr;
			
			
			// Ensure required properties are set
			if ( _videoDevice == null)
				Console.WriteLine( "The video device have not been set. Please set one or both to valid capture devices.\n" );

			// Skip if we are already created
			if ( (int)_graphState < (int)GraphState.SetParamDevice)
			{
				// Garbage collect, ensure that previous filters are released
				GC.Collect();
				
				GetInterfaces();
				


				// Get the video device and add it to the filter graph
				if ( VideoDevice != null )
				{
					_devFilter = (IBaseFilter) Marshal.BindToMoniker( VideoDevice.MonikerString );
				}

				// Retrieve the stream control interface for the video device
				// FindInterface will also add any required filters
				// (WDM devices in particular may need additional
				// upstream filters to function).

				// Try looking for an interleaved media type
				object comObj = null;
				cat  = PinCategory.Capture;
				med = MediaType.Interleaved;
				Guid iid = typeof(IAMStreamConfig).GUID;

				hr = _capGraph.FindInterface( ref cat, ref med,_devFilter, ref iid, out comObj );
				if( hr != 0 )
				{
					// If not found, try looking for a video media type
					med = MediaType.Video;
					hr = _capGraph.FindInterface( ref cat, ref med, _devFilter, ref iid, out comObj );
				
					if( hr != 0 )
						comObj= null;
				}

				//Instead of calling OleCreatePropertyFrame do ...
				_videoStreamConfig=comObj as IAMStreamConfig;

				if(_videoStreamConfig == null) 
				{
					Console.WriteLine("IAMStreamConfig not present");
				}

			
				
				// Add the graph to the Running Object Table so it can be
				// viewed with GraphEdit
				SetupGraph(_graphState);

				#if DEBUG
					DsROT.AddGraphToRot( _graphBuilder, out _rotCookie );
				#endif	

				// Reload any video crossbars
				if ( _videoSources != null ) _videoSources.Dispose(); _videoSources = null;

				// Reload any property pages exposed by filters
				if ( _propertyPages != null ) _propertyPages.Dispose(); _propertyPages = null;

				// Reload capabilities of video device
				_videoCaps = null;

				_graphState = GraphState.Created;
			}
			else if( (int)_graphState < (int)GraphState.Created)
			{
				Type comtype = null;
				object comobj = null;
				// Get a new Instance of InfTee when Resolution or Framerate changed
				_infTee = null;
				// Get the Infinite Tee Pin Filter
				comtype = Type.GetTypeFromCLSID( Clsid.InfTee);
				if( comtype == null )
					throw new NotSupportedException( "DirectX (8.1 or higher) not installed?" );
				comobj = Activator.CreateInstance(comtype);
				Console.WriteLine("Infinite Pin Tee Filter wurde erzeugt");
				_infTee = (IBaseFilter) comobj; comobj = null;
				// Add the graph to the Running Object Table so it can be
				// viewed with GraphEdit
				SetupGraph(_graphState);
				_graphState = GraphState.Created;
			}
				
		}

		/// <summary>
		///  Connects the filters of a previously created graph 
		///  (created by createGraph()). Once rendered the graph
		///  is ready to be used. This method may also destroy
		///  streams if we have streams we no longer want.
		/// </summary>
		protected virtual void renderGraph()
		{
//			Guid					cat;
//			Guid					med;
			int						hr;
			int						nextPin;
			bool					didSomething = false;
			IPin					pinIn;
			IPin					pinOut;
			const int WS_CHILD			= 0x40000000;	
			const int WS_CLIPCHILDREN	= 0x02000000;
			const int WS_CLIPSIBLINGS	= 0x04000000;

			assertStopped();

			// Ensure required properties set
			if ( _filename == null )
				throw new ArgumentException( "The _filename property has not been set to a file.\n" );

			// Stop the graph
			if ( _mediaCtrl != null )
				_mediaCtrl.Stop();

			// Create the graph if needed (group should already be created)
			createGraph();



			// Derender the graph if we have a capture or preview stream
			// that we no longer want. We can't derender the capture and 
			// preview streams seperately. 
			// Notice the second case will leave a capture stream intact
			// even if we no longer want it. This allows the user that is
			// not using the preview to Stop() and Start() without
			// rerendering the graph.
			if ( !_wantPreviewRendered && _isPreviewRendered )
				derenderGraph();
			if ( !_wantCaptureRendered && _isCaptureRendered )
				if ( _wantPreviewRendered )
					derenderGraph();


			// Render capture stream (only if necessary)
			if ( _wantCaptureRendered && !_isCaptureRendered )
			{
//				// Render the file writer portion of graph (mux -> file)
//				Guid mediaSubType = MediaSubType.Avi;
//				hr = _capGraph.SetOutputFileName( ref mediaSubType, _filename, out _muxFilter, out _fileWriterFilter );
//				if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );

				// Render video (video -> mux)
				if ( VideoDevice != null )
				{
					//***********************
					// MUX
					//***********************
					hr = DsUtils.NextPin(_infTee,PinDirection.Output,out nextPin);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);

					hr= DsUtils.GetPin(_infTee,PinDirection.Output,nextPin-1, out pinOut);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR(hr);			
			
					Guid sub = MediaSubType.Avi;
					hr = _capGraph.SetOutputFileName( ref sub, _filename, out _muxFilter, out _fileWriterFilter );
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );					

					if(VideoCompressor == null )
					{
						hr= DsUtils.GetPin(_muxFilter,PinDirection.Input,0, out pinIn);
						if( hr < 0 )
							Marshal.ThrowExceptionForHR(hr);

						hr = _graphBuilder.Connect(pinOut,pinIn);
						if( hr < 0 )
							Marshal.ThrowExceptionForHR( hr );
					}
					else
					{
						//_videoCompressorFilter = (IBaseFilter) Marshal.BindToMoniker( VideoCompressor.MonikerString ); 
						hr = _graphBuilder.AddFilter( _videoCompressorFilter, "Video Compressor" );
						if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );
				
						DsUtils.GetPin(_videoCompressorFilter,PinDirection.Input,0,out pinIn);
						if( hr < 0 )
							Marshal.ThrowExceptionForHR(hr);

						hr = _graphBuilder.Connect(pinOut,pinIn);
						if( hr < 0 )
							Marshal.ThrowExceptionForHR( hr );

						DsUtils.GetPin(_videoCompressorFilter,PinDirection.Output ,0,out pinOut);
						if( hr < 0 )
							Marshal.ThrowExceptionForHR(hr);

						hr= DsUtils.GetPin(_muxFilter,PinDirection.Input,0, out pinIn);
						if( hr < 0 )
							Marshal.ThrowExceptionForHR(hr);

						hr = _graphBuilder.Connect(pinOut,pinIn);
						if( hr < 0 )
							Marshal.ThrowExceptionForHR( hr );
					}



					//***********************
					// END MUX
					//***********************
					Console.WriteLine("Aktueller Aufbau des Filter Graphs:");
					hr = DsUtils.EnumFilters(_filterGraph);
					if( hr < 0 )
						Marshal.ThrowExceptionForHR( hr );
//					// Try interleaved first, because if the device supports it,
//					// it's the only way to get audio as well as video
//					cat = PinCategory.Capture;
//					med = MediaType.Interleaved;
//					hr = _capGraph.RenderStream( ref cat, ref med, _devFilter, _videoCompressorFilter, _muxFilter ); 
//					if( hr < 0 ) 
//					{
//						med = MediaType.Video;
//						hr = _capGraph.RenderStream( ref cat, ref med, _devFilter, _videoCompressorFilter, _muxFilter ); 
//						if ( hr == -2147220969 ) throw new DeviceInUseException( "Video device", hr );
//						if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );
//					}
				}

				_isCaptureRendered = true;
				didSomething = true;
			}

			// Render preview stream (only if necessary)
			if ( _wantPreviewRendered && !_isPreviewRendered )
			{
//				// Render preview (video -> renderer)
//				cat = PinCategory.Preview;
//				med = MediaType.Video;
//				hr = _capGraph.RenderStream( ref cat, ref med, _devFilter, null, null ); 
//				if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );

				//*************************************************
				// Preview Filter
				//*************************************************
				hr = DsUtils.NextPin(_infTee,PinDirection.Output,out nextPin);
				if( hr < 0 )
					Marshal.ThrowExceptionForHR(hr);
				hr= DsUtils.GetPin(_infTee,PinDirection.Output,nextPin-1, out pinOut);
				if( hr < 0 )
					Marshal.ThrowExceptionForHR(hr);
			
				hr =_graphBuilder.Render(pinOut);
				if( hr < 0)
					Marshal.ThrowExceptionForHR( hr );
				//*************************************************
				// END Preview Filter
				//*************************************************
				Console.WriteLine("Aktueller Aufbau des Filter Graphs:");
				hr = DsUtils.EnumFilters(_filterGraph);
				if( hr < 0 )
					Marshal.ThrowExceptionForHR( hr );

				// Get the IVideoWindow interface
				_videoWin = (IVideoWindow) _graphBuilder;

				// Set the video window to be a child of the main window
				hr = _videoWin.put_Owner( _previewWindow.Handle );
				if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );

				// Set video window style
				hr = _videoWin.put_WindowStyle( WS_CHILD | WS_CLIPCHILDREN | WS_CLIPSIBLINGS);
				if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );

				// Use helper function to position video window in client rect of owner window
				//resizeVideoWindow();

				// Position video window in client rect of owner window
				_previewWindow.Resize += new EventHandler( onPreviewWindowResize );
				onPreviewWindowResize( this, null );

				// Make the video window visible, now that it is properly positioned
				hr = _videoWin.put_Visible( DsHlp.OATRUE );
				if( hr < 0 ) Marshal.ThrowExceptionForHR( hr );

				_isPreviewRendered = true;
				didSomething = true;
			}

			if ( didSomething )
				_graphState = GraphState.Rendered;
		}

		/// <summary>
		///  Setup and start the preview window if the user has
		///  requested it (by setting _previewWindow).
		/// </summary>
		protected void startPreviewIfNeeded()
		{
			// Render preview 
			//if ( _wantPreviewRendered && _isPreviewRendered && !_isCaptureRendered )
			
			if ( !_wantCapture && !_isCaptureRendered )
			{
				// Run the graph (ignore errors)
				// We can run the entire graph becuase the capture
				// stream should not be rendered (and that is enforced
				// in the if statement above)
				_mediaCtrl.Run();
			}
		}

		/// <summary>
		///  Disconnect and remove all filters except the device
		///  and compressor filters. This is the opposite of
		///  renderGraph(). Soem properties such as FrameRate
		///  can only be set when the device output pins are not
		///  connected. 
		/// </summary>
		protected virtual void derenderGraph()
		{
			// Stop the graph if it is running (ignore errors)
			if ( _mediaCtrl != null )
				_mediaCtrl.Stop();

			// Free the preview window (ignore errors)
			if ( _videoWin != null )
			{
				_videoWin.put_Visible( DsHlp.OAFALSE );
				_videoWin.put_Owner( IntPtr.Zero );
				_videoWin = null;
			}

			// Remove the Resize event handler
			if ( _previewWindow != null )
				_previewWindow.Resize -= new EventHandler( onPreviewWindowResize );

			if ( (int)_graphState >= (int)GraphState.Rendered )
			{
				// Update the state
				_graphState = GraphState.Created;
				_isCaptureRendered = false;
				_isPreviewRendered = false;

				// Disconnect all filters downstream of the 
				// video and audio devices. If we have a compressor
				// then disconnect it, but don't remove it
//				if ( _infTee != null )
//					removeDownstream( _infTee, (_videoCompressor==null) );
				if ( _devFilter != null )
				{
					removeDownstream( _devFilter, (_videoCompressor==null) );
					
					if(!_wantCapture)
					{
						Type comtype = null;
						object comobj = null;
						// Get a new Instance of InfTee when Resolution or Framerate changed
						_infTee = null;
						// Get the Infinite Tee Pin Filter
						comtype = Type.GetTypeFromCLSID( Clsid.InfTee);
						if( comtype == null )
							throw new NotSupportedException( "DirectX (8.1 or higher) not installed?" );
						comobj = Activator.CreateInstance(comtype);
						Console.WriteLine("Infinite Pin Tee Filter wurde erzeugt");
						_infTee = (IBaseFilter) comobj; comobj = null;
						SetupGraph(_graphState);
						_graphState = GraphState.Rendered;
					}
					else
					{
						_graphState = GraphState.Rendered;
					}
				}

				// These filters should have been removed by the
				// calls above. (Is there anyway to check?)
				_muxFilter = null;
				_fileWriterFilter = null;
			}
		}

		/// <summary>
		///  Removes all filters downstream from a filter from the graph.
		///  This is called only by derenderGraph() to remove everything
		///  from the graph except the devices and compressors. The parameter
		///  "removeFirstFilter" is used to keep a compressor (that should
		///  be immediately downstream of the device) if one is begin used.
		/// </summary>
		protected void removeDownstream( IBaseFilter filter, bool removeFirstFilter )
		{
			int hr;

//			Console.WriteLine("Aktueller Aufbau vor removeDownstream des Filter Graphs:");
//			int hr = DsUtils.EnumFilters(_filterGraph);
//			if( hr < 0 )
//				Marshal.ThrowExceptionForHR( hr );

			// Get a pin enumerator off the filter
			IEnumPins pinEnum;
			hr = filter.EnumPins( out pinEnum );
			pinEnum.Reset();



			if( (hr == 0) && (pinEnum != null) )
			{
				// Loop through each pin
				IPin[] pins = new IPin[1];
				int f;
				do
				{
					// Get the next pin
					hr = pinEnum.Next( 1, pins, out f );
					if( (hr == 0) && (pins[0] != null) )
					{
						// Get the pin it is connected to
						IPin pinTo = null;
						pins[0].ConnectedTo( out pinTo );
						if ( pinTo != null )
						{
							// Is this an input pin?
							PinInfo info = new PinInfo();
							hr = pinTo.QueryPinInfo( out info );
							if( (hr == 0) && (info.dir == (PinDirection.Input)) )
							{
								// Recurse down this branch
								removeDownstream( info.filter, true );

								// Disconnect 
								_graphBuilder.Disconnect( pinTo );
								_graphBuilder.Disconnect( pins[0] );

								// Remove this filter
								// but don't remove the video or audio compressors
								//( info.filter != _videoCompressorFilter ) &&
								if ( ( info.filter != _audioCompressorFilter ) )
									_graphBuilder.RemoveFilter( info.filter );
							}
							Marshal.ReleaseComObject( info.filter );
							Marshal.ReleaseComObject( pinTo );
						}
						Marshal.ReleaseComObject( pins[0] );
					}
				}
				while( hr == 0 );

//				Console.WriteLine("Aktueller Aufbau nach removeDownstream des Filter Graphs:");
//				hr = DsUtils.EnumFilters(_filterGraph);
//				if( hr < 0 )
//					Marshal.ThrowExceptionForHR( hr );
				Marshal.ReleaseComObject( pinEnum ); pinEnum = null;
			}
		}

		/// <summary>
		///  Completely tear down a filter graph and 
		///  release all associated resources.
		/// </summary>
		protected void destroyGraph()
		{
			// Derender the graph (This will stop the graph
			// and release preview window. It also destroys
			// half of the graph which is unnecessary but
			// harmless here.) (ignore errors)
			try{ derenderGraph(); } 
			catch {}

			// Update the state after derender because it
			// depends on correct status. But we also want to
			// update the state as early as possible in case
			// of error.
			_graphState = GraphState.Null;
			_isCaptureRendered = false;
			_isPreviewRendered = false;

			// Remove graph from the ROT
			if ( _rotCookie != 0 )
			{
				DsROT.RemoveGraphFromRot( ref _rotCookie );
				_rotCookie = 0;
			}

			// Remove filters from the graph
			// This should be unnecessary but the Nvidia WDM
			// video driver cannot be used by this application 
			// again unless we remove it. Ideally, we should
			// simply enumerate all the filters in the graph
			// and remove them. (ignore errors)
			IBaseFilter remove =null;

			if ( _muxFilter != null )
				_graphBuilder.RemoveFilter( _muxFilter );
			if ( _videoCompressorFilter != null )
				_graphBuilder.RemoveFilter( _videoCompressorFilter  );
			if ( _audioCompressorFilter != null )
				_graphBuilder.RemoveFilter( _audioCompressorFilter  );
			if ( _nullRenderer != null )
				_graphBuilder.RemoveFilter( _nullRenderer );
			if ( _nullRenderer2 != null )
				_graphBuilder.RemoveFilter( _nullRenderer2);
			if ( _sampGrabber != null )
				remove = _sampGrabber as IBaseFilter;
				_graphBuilder.RemoveFilter(remove );
			if ( _transformFilter != null )
				_graphBuilder.RemoveFilter( _transformFilter );
			if ( _infTee != null )
				_graphBuilder.RemoveFilter( _infTee );

			if ( _devFilter != null )
				_graphBuilder.RemoveFilter( _devFilter );
			if ( _audioDeviceFilter != null )
				_graphBuilder.RemoveFilter( _audioDeviceFilter );
	


			// Clean up properties
			if ( _videoSources != null )
				_videoSources.Dispose(); _videoSources = null;
			if ( _audioSources != null )
				_audioSources.Dispose(); _audioSources = null;
			if ( _propertyPages != null )
				_propertyPages.Dispose(); _propertyPages = null;
			if ( _tuner != null )
				_tuner.Dispose(); _tuner = null;

			// Cleanup
			if ( _graphBuilder != null )
				Marshal.ReleaseComObject( _graphBuilder );  _graphBuilder = null;
			if ( _capGraph != null )
				Marshal.ReleaseComObject( _capGraph ); _capGraph = null;
			if ( _muxFilter != null )
				Marshal.ReleaseComObject( _muxFilter ); _muxFilter = null;
			if ( _fileWriterFilter != null )
				Marshal.ReleaseComObject( _fileWriterFilter ); _fileWriterFilter = null;
			if ( _devFilter != null )
				Marshal.ReleaseComObject( _devFilter ); _devFilter = null;
			if ( _audioDeviceFilter != null )
				Marshal.ReleaseComObject( _audioDeviceFilter ); _audioDeviceFilter = null;
			if ( _videoCompressorFilter != null )
				Marshal.ReleaseComObject( _videoCompressorFilter ); _videoCompressorFilter = null;
			if ( _audioCompressorFilter != null )
				Marshal.ReleaseComObject( _audioCompressorFilter ); _audioCompressorFilter = null;
			if( _sampGrabber != null )
				Marshal.ReleaseComObject( _sampGrabber ); _sampGrabber = null;
			if( _infTee != null )
				Marshal.ReleaseComObject( _infTee ); _infTee = null;
			if( _nullRenderer != null )
				Marshal.ReleaseComObject( _nullRenderer ); _nullRenderer = null;
			if( _nullRenderer2 != null )
				Marshal.ReleaseComObject( _nullRenderer2 ); _nullRenderer2 = null;
			if( _transformFilter != null )
				Marshal.ReleaseComObject( _transformFilter ); _transformFilter = null;

			// These are copies of _graphBuilder
			_mediaCtrl = null;
			_videoWin = null;
			_baseGrabFlt = null;
			_filterGraph = null;

			// For unmanaged objects we haven't released explicitly
			GC.Collect();
		}

		/// <summary> Resize the preview when the _previewWindow is resized </summary>
		protected void onPreviewWindowResize(object sender, EventArgs e)
		{
			if ( _videoWin != null )
			{
				// Position video window in client rect of owner window
				Rectangle rc = _previewWindow.ClientRectangle;
				_videoWin.SetWindowPosition( 0, 0, rc.Right, rc.Bottom );
			}
		}
	
		/// <summary> 
		///  Get a valid temporary _filename (with path). We aren't using 
		///  Path.GetTempFileName() because it creates a 0-byte file 
		/// </summary>
		protected string getTempFilename()
		{
			string s;
			try
			{
				int count = 0;
				int i;
				Random r = new Random();
				string tempPath = Path.GetTempPath();
				do
				{
					i = r.Next();
					s = Path.Combine( tempPath, i.ToString("X") + ".avi" );
					count++;
					if ( count > 100 ) throw new InvalidOperationException( "Unable to find temporary file." );
				} while ( File.Exists( s ) );
			}
			catch { s = "c:\temp.avi"; }
			return( s );
		}

		/// <summary>
		///  Retrieves the value of one member of the IAMStreamConfig format block.
		///  Helper function for several properties that expose
		///  video/audio settings from IAMStreamConfig.GetFormat().
		///  IAMStreamConfig.GetFormat() returns a AMMediaType struct.
		///  AMMediaType.formatPtr points to a format block structure.
		///  This format block structure may be one of several 
		///  types, the type being determined by AMMediaType.formatType.
		/// </summary>
		protected object getStreamConfigSetting( IAMStreamConfig streamConfig, string fieldName)
		{
			if ( streamConfig == null )
				throw new NotSupportedException();
			assertStopped();
//			derenderGraph();

			object returnValue = null;
			IntPtr pmt = IntPtr.Zero;
			AMMediaType mediaType = new AMMediaType();

			try 
			{
				// Get the current format info
				int hr = streamConfig.GetFormat( out mediaType );
				if ( hr != 0 )
					Marshal.ThrowExceptionForHR( hr );
				//Marshal.PtrToStructure( pmt, mediaType );

				// The formatPtr member points to different structures
				// dependingon the formatType
				object formatStruct;
				if ( mediaType.formatType == FormatType.WaveEx )
					formatStruct = new WaveFormatEx();
				else if ( mediaType.formatType == FormatType.VideoInfo )
					formatStruct = new VideoInfoHeader();
				else if ( mediaType.formatType == FormatType.VideoInfo2 )
					formatStruct = new VideoInfoHeader2();
				else
					throw new NotSupportedException( "This device does not support a recognized format block." );

				// Retrieve the nested structure
				Marshal.PtrToStructure( mediaType.formatPtr, formatStruct );

				// Find the required field
				Type structType = formatStruct.GetType();
				FieldInfo fieldInfo = structType.GetField( fieldName );
				if ( fieldInfo == null )
					throw new NotSupportedException( "Unable to find the member '" + fieldName + "' in the format block." );

				// Extract the field's current value
				returnValue = fieldInfo.GetValue( formatStruct ); 
						
			}
			finally
			{
				DsUtils.FreeAMMediaType( mediaType );
				Marshal.FreeCoTaskMem( pmt );
			}
			renderGraph();
			startPreviewIfNeeded();

			return( returnValue );
		}

		/// <summary>
		///  Set the value of one member of the IAMStreamConfig format block.
		///  Helper function for several properties that expose
		///  video/audio settings from IAMStreamConfig.GetFormat().
		///  IAMStreamConfig.GetFormat() returns a AMMediaType struct.
		///  AMMediaType.formatPtr points to a format block structure.
		///  This format block structure may be one of several 
		///  types, the type being determined by AMMediaType.formatType.
		/// </summary>
		protected object setStreamConfigSetting( IAMStreamConfig streamConfig, string fieldName, object newValue)
		{
			if ( streamConfig == null )
				throw new NotSupportedException();
			assertStopped();
			derenderGraph();

			object returnValue = null;
			IntPtr pmt = IntPtr.Zero;
			AMMediaType mediaType = new AMMediaType();

			try 
			{
				// Get the current format info
				int hr = streamConfig.GetFormat( out mediaType );
				if ( hr != 0 )
					Marshal.ThrowExceptionForHR( hr );
				//Marshal.PtrToStructure( pmt, mediaType );

				// The formatPtr member points to different structures
				// dependingon the formatType
				object formatStruct;
				if ( mediaType.formatType == FormatType.WaveEx )
					formatStruct = new WaveFormatEx();
				else if ( mediaType.formatType == FormatType.VideoInfo )
					formatStruct = new VideoInfoHeader();
				else if ( mediaType.formatType == FormatType.VideoInfo2 )
					formatStruct = new VideoInfoHeader2();
				else
					throw new NotSupportedException( "This device does not support a recognized format block." );

				// Retrieve the nested structure
				Marshal.PtrToStructure( mediaType.formatPtr, formatStruct );

				// Find the required field
				Type structType = formatStruct.GetType();
				FieldInfo fieldInfo = structType.GetField( fieldName );
				if ( fieldInfo == null )
					throw new NotSupportedException( "Unable to find the member '" + fieldName + "' in the format block." );

				// Update the value of the field
				fieldInfo.SetValue( formatStruct, newValue );

				// PtrToStructure copies the data so we need to copy it back
				Marshal.StructureToPtr( formatStruct, mediaType.formatPtr, false ); 

				// Save the changes
				hr = streamConfig.SetFormat( mediaType );
				if ( hr != 0 )
					Marshal.ThrowExceptionForHR( hr );
			}
			finally
			{
				DsUtils.FreeAMMediaType( mediaType );
				Marshal.FreeCoTaskMem( pmt );
			}
			renderGraph();
			startPreviewIfNeeded();

			return( returnValue );
		}

		/// <summary>
		///  Assert that the class is in a Stopped state.
		/// </summary>
		protected void assertStopped()
		{
			if ( !Stopped )
				throw new InvalidOperationException( "This operation not allowed while Capturing. Please Stop the current capture." ); 
		}
	}


}
