using System;
using System.Collections;
using MT.Tracker.DataStructures;


namespace MT.Tracker
{
	public class HypothesisSuperSet: HypothesisSet
	{
		SortedList _f1 = new SortedList();
		SortedList _f2 = new SortedList();
		SortedList _f3 = new SortedList();

		int _f1Neutral = 0;
		int _f2Neutral = 0;
		int _f3Neutral = 0; //This one will always be zero, its only here for consistency

		new public void Add(Hypothesis H)
		{	
			FeaturePoint P;
			SortedList f;

			//The Idea is, to sort all Hypothesises 3 times for each frame once, according to which 
			//Featurepoints they use. This will help us later, when we want to detect the best cost collection
			//of Hypothesises which have no collisions with each other (they don't use the same points)

			//In the End we have a Collection(Sortedlist) for each Frame in which we store SubCollections
			//Of Hypothesises which use different FPs of that frame - so all Hypothesises which use
			//the same Featurepoint in f are also in the Same SubCollection hence we can avoid
			//collisions by choosing only Hypothesis Groups which are never in the same SubCollection over
			//all 3 frames

			base.Add(H);

			P = H.P1;
			f = _f1;
			Analyse(f,H,P,ref _f1Neutral);

			P = H.P2;
			f = _f2;
			Analyse(f,H,P,ref _f2Neutral);

			P = H.P3;
			f = _f3;
			Analyse(f,H,P,ref _f3Neutral);

		}

		/// <summary>
		/// Analyses a given Featurepoint P of a Hypothesis H and sorts it into a HypothesisSet of the Hashtable f.
		/// </summary>
		/// <param name="f">SortedList which contains HypothesisSets for each Point P</param>
		/// <param name="H">Hypothesis to sort.</param>
		/// <param name="P">Point of the Hypothesis which is to analyse.</param>
		private void Analyse(SortedList f, Hypothesis H, FeaturePoint P, ref int NeutralCount)
		{
			HypothesisSet HS;

			//We want to put the Hypothesis H into a Set from f, 
			//according to the Point P they use - so that all Hs which
			//use the same Point P are in the Same Set


			//Only P's which are not doublelinked Point are put into P Groups
			//All others are sorted into a Neutral Group
			//because they are already completely linked and 
			//therefor cannot cause any collisions (Tracker takes care of that)
//			if (P.LinkType == eFPLinkType.Zero)
			if (P.LinkType != eFPLinkType.DoubleLinked)

			{	
				//Put zerolinked Points into their P-Group
				//if that group doesn't exist yet, create it
				if (!f.Contains(P))
				{
					HS = new HypothesisSet();
					f.Add(P,HS);
				}
				else
				{
					HS = (HypothesisSet)f[P];
				}
				HS.Add(H);

			}
			else
				NeutralCount++;

			//			else
			//			{	
			//				//Put not zerolinked Point into the "neutralSet"
			//				//Group - create it, if it doesn't exist yet.
			//				if (!f.Contains("neutralSet"))
			//				{
			//					HS = new HypothesisSet();
			//					f.Add("neutralSet",HS);
			//				}
			//				else
			//				{
			//					HS = (HypothesisSet)f["neutralSet"];
			//				}
			//				HS.Add(H);
			//			}


		}

		/// <summary>
		/// Returns a Set of not-colliding Hypothesises
		/// which has been determined to have the lowest
		/// overall Costs
		/// </summary>
		/// <returns></returns>
		public HypothesisSet getBestSet()
		{
			//Console.WriteLine(base.Count);
			//Autodeactivate Tracking Process in case of too much Hypothesises
			//to prevent blocking of Application
			//ToDo: call an easier Version of getBestSet so that at least a simple tracking
			//exists
			if (base.Count > 30)
				return new HypothesisSet();

			int i;
			int[,] i3 = new int[_f3.Count,2];		//First Column stores the currently selected SubCollection, 
			bool[] i3active = new bool[_f3.Count];	/*Tells, which of the currently selected Combination of SubCollections is
													 *active in case there is a lower number of free points in F1 and F2
													 *than in F3 */

			int	_countvalids = 0;					//counts valid sets and stops after reaching maxvalids
			int _maxvalids = 5;
			//Maximum number of allowed Hypothesises in a BestSet is
			//the minimal number of FPs in one of the Frames
			int minimumFP = Math.Min(Math.Min(_f1.Count+_f1Neutral,_f2.Count+_f2Neutral),_f3.Count+_f3Neutral);

			//Initalise First active Set of SubCollections
			for(i = 0;i<minimumFP;i++)
				i3active[i] = true;

			for(i = minimumFP;i<_f3.Count;i++)
				i3active[i] = false;



			
			//Second Column stores the number of Members of that SubCollection
			bool[] i2new = new bool[_f2.Count];
			bool[] i1new = new bool[_f1.Count];
			bool[] i2;
			bool[] i1;

			bool Collision = false;

			HypothesisSet TestSet;
			HypothesisSet BestSet = null;

			//Initialise TestHypothesis Array
			for (i=0;i<i3.GetLength(0);i++)
			{
				i3[i,0]=0;
				i3[i,1]=((HypothesisSet)_f3.GetByIndex(i)).Count;
			}



			do
			{
				do
				{
					//Reset Collisioncheckers
					i2 = (bool[])i2new.Clone();
					i1 = (bool[])i1new.Clone();

//					for (i=0;i<i1.Length;i++)
//					{
//						i1[i] = false;
//					}
//
//					for (i=0;i<i2.Length;i++)
//					{
//						i2[i] = false;
//					}

					TestSet = new HypothesisSet();

					//We create a Testset by starting from the SubCollections of 
					//Frame 3, since every Hypothesis contains one zerolinked point of 
					//Frame 3.
					//We start out by picking the first Element of each Subcollections
					//and start increasing the indices so that every possible combination 
					//is formed

					//fill up testset according to i3 indices
					// SH changed 8.11:
					//	Added validity check for each new added hypothesis instead
					//  of one check at the end - this will improve speed by a great
					//  deal.
					for (i=0;i<i3.GetLength(0) && !Collision;i++)
					{
						if (i3active[i])
							Collision = AddIfValid(TestSet,((HypothesisSet)_f3.GetByIndex(i)).GetByIndex(i3[i,0]),ref i1, ref i2);
							//TestSet.Add(((HypothesisSet)_f3.GetByIndex(i)).GetByIndex(i3[i,0]));
					}

					//TODO: has to be deactivated ! The first Testset it automatically the best Set
					if (BestSet == null)
						BestSet = TestSet;

					//Check for collisions
					if (!Collision)
					{	
						//_countvalids++;

						if (BestSet == null)
							BestSet = TestSet;
						else if (TestSet.Cost < BestSet.Cost)
							BestSet = TestSet;
				
					}

					Collision = false;

				}while (incActiveSet(ref i3active));
			}
			while(incTestSet(ref i3) && _countvalids < _maxvalids);


			if (BestSet == null)
				throw new Exception("HypothesisSuperSet.getBestSet(): BestSet was NULL!");
				
			return BestSet;
			
		}


		/// <summary>
		/// Checks if a new Hypothesis doesn't cause a collision and
		/// adds it in that case - otherwise it will not be added and 
		/// the return value will be false.
		/// </summary>
		/// <param name="HS">HypothesisSet to test</param>
		/// <param name="H">Hypothesis to test and add</param>
		/// <param name="i1">Empty Boolean Array</param>
		/// <param name="i2">Empty Boolean Array</param>
		/// <returns></returns>
		private bool AddIfValid(HypothesisSet HS, Hypothesis H,  ref bool[] i1, ref bool[] i2)
		{
			int Index1;
			int Index2;


			//check f1 validity

			//Only doublelinked Points may not collide
			if (H.P1.LinkType != eFPLinkType.DoubleLinked)
			{
				Index1 = _f1.IndexOfKey(H.P1);

				//If the FP of the current H uses
				//an already taken FP we have a collision
				//and the tested set is therefor invalid
				if (i1[Index1])
					return true;

				i1[Index1]=true;
			}

			//check f2 validity

			//Only zerolinked Points may collide
			if (H.P2.LinkType != eFPLinkType.DoubleLinked)
			{
				Index2 = _f2.IndexOfKey(H.P2);

				if (i2[Index2])
					return true;

				i2[Index2]=true;
			
			
			}

			HS.Add(H);

			return false;
		}

		/// <summary>
		/// Outdated - use AddifValid!
		/// Checks if a given Hypothesis Set does not produce any Collisions
		/// </summary>
		/// <param name="HS">HypothesisSet to test</param>
		/// <param name="i1">Empty Boolean Array</param>
		/// <param name="i2">Empty Boolean Array</param>
		/// <returns></returns>
		private bool isValid(HypothesisSet HS, bool[] i1, bool[] i2)
		{
			int Index1;
			int Index2;
			Hypothesis H;


			foreach(DictionaryEntry DE in HS)
			{
				H=((Hypothesis)DE.Value);

				//check f1 validity

				//Only zerolinked Points may collide
				if (H.P1.LinkType == eFPLinkType.Zero)
				{
					Index1 = _f1.IndexOfKey(H.P1);

					//If the FP of the current H uses
					//an already taken FP we have a collision
					//and the tested set is therefor invalid
					if (i1[Index1])
						return false;

					i1[Index1]=true;
				}

				//check f2 validity

				//Only zerolinked Points may collide
				if (H.P2.LinkType == eFPLinkType.Zero)
				{
					Index2 = _f2.IndexOfKey(H.P2);

					if (i2[Index2])
						return false;

					i2[Index2]=true;
				
				}
				
			}

			return true;
		}

		//This Sub allows us to increase the indices of n Subcollections with m(n) Members
		//it will return a TRUE if it has reached the starting position 0,0,0,0... again.
		//The n and m's are stored in IndexArray:
		//  Index    currently selected		last element  
		//  ----------------------------------------------
		//    0				n0					 m0
		//    1				n1					 m1
		//   ...			...					...
		//
		//Example
		//IndexArray:   0, 2	 1, 2	  2, 2	   0, 2     1, 2     2, 2     0, 2           2, 2     0, 2
		//				0, 3 --> 0, 3 --> 0, 3 --> 1, 3 --> 1, 3 --> 1, 3 --> 2, 3 -->...--> 3, 3 --> 0, 3 (Overflow = True)
		//              0, 1     0, 1     0, 1     0, 1     0, 1     0, 1     0, 1           1, 1     0, 1
		//
		public bool incTestSet(ref int[,]IndexArray)
		{
			bool overflow = true;
			int i = 0;

			while(i<IndexArray.GetLength(0) && overflow)
			{
				IndexArray[i,0] = (IndexArray[i,0]+1) % IndexArray[i,1];
				
				if (IndexArray[i,0] != 0)
					overflow = false;

				i++;

			}

			//If we reach an overflow in the very last element, we are at the start again
			if (i==IndexArray.GetLength(0) && overflow == true)
				return false;
			else
				return true;
		}

		//Zusaetzlich zum  Zyklus gibt es einen uebergeordneten AktiverIndexZyklus
		//der alle moeglichen Anordnungen einer bestimmten Menge aktivierter Elemente erlaubt.
		//Nachfolgend ein Beispiel 3 aktiver Elemente in einem Set von 6
		//Total ergeben sich 6 ueber 3 Varianten.
		//Algorithmus: Suche x mit darunterliegendem o - wenn nicht gefunden, fertig
		//wenn gefunden, dann verschiebe x auf o und ordne alle darueberliegenden x
		//auf oberste stellen
		/*	x  x  x  o  x  x  o  x  o  o  x  x  o  x  o  o  x  o  o  o
		 *  x  x  o  x  x  o  x  o  x  o  x  o  x  o  x  o  o  x  o  o
		 *  x  0  x  x  o  x  x  o  o  x  o  x  x  o  o  x  o  o  x  o
		 *  o  x  x  x  o  o  o  x  x  x  o  o  o  x  x  x  o  o  o  x
		 *  o  o  o  o  x  x  x  x  x  x  o  o  o  o  o  o  x  x  x  x
		 *  o  o  o  o  o  o  o  o  o  o  x  x  x  x  x  x  x  x  x  x
		 */
		private bool incActiveSet(ref bool[]ActiveArray)
		{
			bool moved = false;
			int notmoved = 0;
			int i = 0;
			int j;

			while(!moved && i<ActiveArray.GetLength(0)-1)
			{
				//Gesuchtes X gefunden, dann verschieben und
				//alle vorhergehenden X auf Startposition setzen
				if(ActiveArray[i] && !ActiveArray[i+1])
				{	
					ActiveArray[i] = false;
					ActiveArray[i+1]= true;

					//Startpositionen aktivieren
					for(j = 0;j<notmoved;j++)
						ActiveArray[j] = true;

					//Nicht Startpositionen zuruecksetzen
					for(j = notmoved;j<=i;j++)
						ActiveArray[j] = false;

					moved = true;
				}
				else if(ActiveArray[i])
					notmoved++;

				i++;
			}

			//Solange wir was bewegen konnten, ist alles ok
			//Ansonsten haben wir das Ende erreicht.
			if(moved) 
				return true;
			else
				return false;
		}

	}
}
