using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using GraphController;
using DShowNET;
using DShowNET.Device;
using MT.FPSource;
using MT.Network;



namespace MTClient
{
	/// <summary>
	/// Setting Package which the device selector sends to the ClientController to configure its setup
	/// </summary>
//	public struct strSettings
//	{
//		public Filter videoDevice;
//		public enuSWFilterMode SWFilterMode;
//		public enuNetworkMode NetworkMode;
//		public string IP;		//has only to be filled when Network Mode is set to manual.
//	}

	/// <summary>
	/// Used to set, if the client run in
	/// DirectShowTransformFilter Mode or if it runs with the CSharp Feature
	/// Point Extractor
	/// </summary>
	public enum enuSWFilterMode
	{
		DirectShowSWFilter,
		CSharpSWFilter
	}

	/// <summary>
	/// Used to indicate in which State the ClientController runs at the moment
	/// </summary>
	public enum enuControllerState
	{
		DeviceSelector,
		MainWindow
	}

	/// <summary>
	/// Sets, whether the Client uses a Webcam or DVCam Device or if it
	/// receives Videodata from a Filesource
	/// </summary>
	public enum enuCaptureDevice
	{	Webcam,
		DVCam,
		Video
	}

/// <summary>
/// Sets, whether the Client runs in Autoconnect Mode or if an IP is entered Manually
/// </summary>
	public enum enuNetworkMode
	{
		Manual,
		Automatic
	}

	/// <summary>
	/// clsClientController is the Main Controller Class of the MTClient Application.
	/// It contains all Processing layer classes and provides access to them for Formulars and Usercontrols.
	/// </summary>
	public class clsClientController
	{
		private delegate void NewSyncSaveDelegate(int value);
		private delegate void NewRecordSaveDelegate(bool value);

		//Collection of DirectShowFilters
		private Filters _filters = new Filters();

		//FiltergraphController is the Main Controllerclass for all Directshow related operations
		//it is instanted by the DeviceSelector Form when a capturing Device has been selected.
		private FilterGraphController _filterGraphController = null;
		private FilterGraphControllerSource _filterGraphControllerSource =null;

		//SWFilterMode, there are 2 basic modes available: a directshowfilter which
		//sends its data over network communication to this client
		//or a SampleGrabber modul combined with a CSharp Version of the SWFilter
		//that extracts the required Featurepoints and sends it directly to the tracker
		private enuSWFilterMode _eSWFilterMode = enuSWFilterMode.DirectShowSWFilter;

		private enuControllerState	_eControllerState = enuControllerState.DeviceSelector;

		private enuCaptureDevice	_eCaptureDevice = enuCaptureDevice.Webcam;

		private clsFilterServer		_FilterServer;

		private clsCalibration		_Calibration;
		private clsTracker			_Tracker;
		private clsNetworkClient	_NetClient;

		private IFPsource			_FPsource;
//		private IFPsink				_FPsink;

		private Control				_InvokeControl;

		#region "Properties"

		public Filters Filters
		{
			get { return _filters; }
		}

		public clsFilterServer FilterServer
		{
			get { return _FilterServer;}
		}

		public FilterGraphController FilterGraphController
		{
			get { return _filterGraphController; }
		}

		public FilterGraphControllerSource FilterGraphControllerSource
		{
			get { return _filterGraphControllerSource; }
		}

		public clsCalibration Calibration
		{
			get { return _Calibration; }
		}

		public clsNetworkClient NetClient
		{
			get { return _NetClient; }
		}

		public enuSWFilterMode SWFilterMode
		{
			get { return _eSWFilterMode; }
		}

		public enuControllerState State
		{
			get { return _eControllerState; }
		}

		public enuCaptureDevice CaptureDevice
		{
			get { return _eCaptureDevice; }
		}

		public clsTracker Tracker
		{
			get { return _Tracker; }
		}
		#endregion 

		public clsClientController()
		{
			//Setup Network Client
			_NetClient = new clsNetworkClient();
			_NetClient.NewSyncSignal +=new MTClient.clsNetworkClient.NewSyncSignalDelegate(_NetClient_NewSyncSignal);
			_NetClient.Play += new MTClient.clsNetworkClient.PlayDelegate(_NetClient_Play);
			_NetClient.Record += new MTClient.clsNetworkClient.RecordDelegate(_NetClient_Record);

		}

		public Control InvokeControl
		{
			set
			{
				_InvokeControl = value;
			}
		}

		/// <summary>
		/// This Function is called, when the OK Button on the DeviceSelector is pressed and 
		/// when a Video Device was selected.
		/// </summary>
		/// <param name="videoDevice"></param>
		public void SetupDevice(Filter videoDevice, enuSWFilterMode SWFilterMode)
		{
			_filterGraphController = new FilterGraphController( videoDevice, null );

			//Setup(SWFilterMode);
			_FilterServer  = new clsFilterServer(_filterGraphController.swInterface);


			//Set SWFiltermode
			_eSWFilterMode = SWFilterMode;

			switch (_eSWFilterMode)
			{
				case enuSWFilterMode.DirectShowSWFilter:
					
					// Called before starting the main window
					// sets up the Tracking mode for Network based SWFilter
					if (!_FilterServer.Start())
						throw new Exception("Fatal Error: The Filterserver couldn't be started!");
		
					_FPsource = _FilterServer.Server;					
					break;
				case enuSWFilterMode.CSharpSWFilter:

					/// Called before starting Main Window Form - sets the tracking
					/// mode for CSharp based SWFilter
					_filterGraphController.InitializeVideoGrabber();
					_FPsource = _filterGraphController.fpDetector;
					
					break;
			}

			//Setup Calibration
			_Calibration = new clsCalibration(_filterGraphController.PictureGrabber, (clsFPnetSink)_NetClient.FPsink);

			//Setup Tracker
			_Tracker = new clsTracker(_FPsource, _NetClient.FPsink);


			//Set CaptureDeviceMode
			if(_filterGraphController.VideoDevice.Name.StartsWith("Microsoft"))
				_eCaptureDevice = enuCaptureDevice.DVCam;
			else
				_eCaptureDevice = enuCaptureDevice.Webcam;

			//Set State to Controllerstate to fix Filtergraph and SWFilterMode settings
			_eControllerState = enuControllerState.MainWindow;


		}

		/// <summary>
		/// This Function is called, when the OK Button on the DeviceSelector is pressed and 
		/// when a Source File was opened.
		
		/// </summary>
		/// <param name="videoDevice"></param>
		public void SetupSource(string sourceFile, Filter audioDevice, enuSWFilterMode SWFilterMode)
		{
			_filterGraphControllerSource = new FilterGraphControllerSource( sourceFile, audioDevice );
			_filterGraphController = _filterGraphControllerSource;

			//Setup(SWFilterMode);
			_FilterServer  = new clsFilterServer(_filterGraphControllerSource.swInterface);


			//Set SWFiltermode
			_eSWFilterMode = SWFilterMode;

			switch (_eSWFilterMode)
			{
				case enuSWFilterMode.DirectShowSWFilter:
					
					// Called before starting the main window
					// sets up the Tracking mode for Network based SWFilter
					if (!_FilterServer.Start())
						throw new Exception("Fatal Error: The Filterserver couldn't be started!");
		
					_FPsource = _FilterServer.Server;					
					break;
				case enuSWFilterMode.CSharpSWFilter:

					/// Called before starting Main Window Form - sets the tracking
					/// mode for CSharp based SWFilter
					_filterGraphControllerSource.InitializeVideoGrabber();
					_FPsource = _filterGraphControllerSource.fpDetector;
					
					break;
			}

//			//Setup Network Client
//			Old Code
//			_NetClient = new clsNetworkClient();

			//Setup Calibration
			_Calibration = new clsCalibration(_filterGraphControllerSource.PictureGrabber,(clsFPnetSink)_NetClient.FPsink);

			//Setup Tracker
			_Tracker = new clsTracker(_FPsource, _NetClient.FPsink);

			//Set CaptureDeviceMode
			_eCaptureDevice = enuCaptureDevice.Video;
			

			//Set State to Controllerstate to fix Filtergraph and SWFilterMode settings
			_eControllerState = enuControllerState.MainWindow;

		}

		/// This Function is called, to fixate all settings which cannot be changed afterwards.
		private void Setup(enuSWFilterMode SWFilterMode)
		{

		}

		/// <summary>
		/// Call when shutting down the Client Application
		/// this function implements all necessary procedures to properly finish 
		/// all started threads.
		/// </summary>
		public void Shutdown()
		{
//			_filterGraphController.

			if(_FPsource != null)
				_FPsource.Deactivate();

			if(_FilterServer != null)
				_FilterServer.Stop();

			if (_NetClient != null)
				_NetClient.Dispose();

			if(_Tracker != null)
				_Tracker.Dispose();
		}


		/// <summary>
		/// Receives Serverside SyncSignal and sets it in the SWFilter
		/// </summary>
		/// <param name="Value"></param>
		private void _NetClient_NewSyncSignal(int Value)
		{  
			_InvokeControl.BeginInvoke(new NewSyncSaveDelegate(_filterGraphController.SetSync),new Object[] {Value});
		}

		/// <summary>
		/// Receives Serverside StopPlay Signal and sets it in the UserControl
		/// </summary>
		/// <param name="Value"></param>
		private void _NetClient_Play(bool enable)
		{  
			if(enable)
			{
				_filterGraphControllerSource.ClipPlay();
			}
			else
			{
				_filterGraphControllerSource.ClipStop();
			}

		}

		private void _NetClient_Record(bool enable)
		{  

			_InvokeControl.BeginInvoke(new NewRecordSaveDelegate(RecordSave),new Object[] {enable});

		}

		private void RecordSave (bool enable)
		{
			if(enable)
			{
				try
				{
					if ( _filterGraphController == null )
						throw new ApplicationException( "Please select a video and/or audio device." );
					if ( !_filterGraphController.Cued )
						_filterGraphController.Filename = @"C:\tracking.avi";
					_filterGraphController.Start();
			
				}
				catch (Exception ex)
				{
					MessageBox.Show( ex.Message + "\n\n" + ex.ToString() );
				}
			}
			else
			{
				try
				{
					if ( _filterGraphController== null )
						throw new ApplicationException( "Please select a video and/or audio device." );
					_filterGraphController.Stop();
				}
				catch (Exception ex)
				{
					MessageBox.Show( ex.Message + "\n\n" + ex.ToString() );
				}
			}
		}

	}
}
