using System;
using System.Collections;
using System.Windows.Forms;
using System.Threading;

using MT.Tracker;
using MT.Tracker.DataStructures;
using MT.Network;
using GraphController;

namespace MTClient
{
	/// <summary>
	/// clsTracker contains all operations concerning the dataflow for the tracker
	/// this includes a Featurepoint Source Object (_FPSource) which receives the
	/// data from our Filtergraph, 3 tracker instances (one for each RGB Color) and 
	/// finally the FeaturepointSink object to which we send the date for further
	/// processing - especially the final 3D Tracking
	///  
	/// Use: Create Tracker and call Initialize to startup all internal Controls.
	///		 Of course you could put all the initalize commands directly into the constructor
	///		 but then you would miss some important events, which are connected after the constructor
	///		 was already called.
	/// </summary>
	public class clsTracker: IDisposable
	{

		//declare Event Delegates
		public delegate void NewFrameDelegate(FeaturePointData FPD);
		public delegate void RefreshFPSDelegate(int FPS);

		//declare Events
		public event NewFrameDelegate NewFrame ;
		public event RefreshFPSDelegate	RefreshFPS;
//		public event NewHypothesisesDelegate NewHypothesises;

		private Thread					_thrFPS;
		private int						_FPS;

		private	IFPsource				_Source;
		private IFPsink					_Sink;

//		private	Tracker					_TrackerR;	//Contains a Tracker for each color
//		private	Tracker					_TrackerG;	
//		private	Tracker					_TrackerB;	
		private	RGBTracker				_Tracker;
		private	FeaturePointFrame		_FrameOld;	//Contains 1 Frameset of Featurepoints
		private	FeaturePointFrame		_FrameCur;	
		private	FeaturePointFrame		_FrameNew;	
		private	FeaturePointData		_FrameBuffer; //Contains a given amount of already tracked frames
		private	int						_buffersize = 15;	//Defines the maximum size of the buffer

		private		bool				_TrackerInitialized = false;
		private		bool				_DoubleFrameCheck = true;

		#region "Properties"

		public bool DoubleFrameCheck
		{	
			get { return _DoubleFrameCheck; }
			set { _DoubleFrameCheck = value; }
		}

		public int FrameBufferSize
		{
			get { return _buffersize; }
			set { _buffersize = value; }
		}

		public RGBTracker	Tracker
		{
			get { return _Tracker; }
		}

		#endregion

		//Connects a IPAN Tracker between a IFPsource and an IFPsink
		public clsTracker(IFPsource Source, IFPsink Sink)
		{	//Set Featurepoint Source and connect its NewFrame event.
			_Source = Source;
			_Source.NewFrame +=new GraphController.NewFrameDelegate(ReceivedNewFrame);

			_Sink = Sink;

			//Create Tracker instantes
			_Tracker = new RGBTracker();

			_FrameBuffer = new FeaturePointData("");

			_thrFPS = new Thread(new ThreadStart(FPSRefresher));
			_thrFPS.Start();
		}

		public void DeactivateSource()
		{ 
			_Source.Deactivate();
		}

		public void ActivateSource()
		{
			_Source.Activate();
		}


		int _testcounter=0;

		/// <summary>
		/// New Frame Eventhandler - is called by whomever sends us a new Frame
		/// </summary>
		/// <param name="FPframe">Datastructure containing all Featurepoints of one Frame</param>
		protected void ReceivedNewFrame(FeaturePointFrame FPframe)
		{	
			//Actualize Framepointers
			_FrameOld = _FrameCur;
			_FrameCur = _FrameNew;
			_FrameNew = FPframe;

			//Actualize Buffer
			_FrameBuffer.AddFrame(FPframe.fid,FPframe);
			if (_FrameBuffer.CountFrames()>_buffersize)
				_FrameBuffer.RemoveFirstFrame();

			//Track for each color
			Track();
			
			//Send Data to next destination
			if (_FrameOld!=null)
			{
//				if (_FrameCur!=null)
//					//					if (_FrameOld.Sync != _FrameCur.Sync)
//				{
////					Console.Write("test \r\n");
////					Console.Write("\fformfeed test\r\n");
////					Console.Write("blaaaaaaaaaaaaaaa\r\n");
////					Console.Write("\r\r\rblaaaaaaaaaaaaaaa\r\n");
//					Console.Write(_FrameOld.Sync.ToString() + " \f " + _testcounter.ToString() + "\r");
//				}
				_Sink.Send(_FrameOld);
			}

			_testcounter++;

			//Raise Event to announce new tracked Frame
			if (NewFrame != null)
				foreach (Delegate d in NewFrame.GetInvocationList())
				{
					Control c = d.Target as Control;
					if(c!=null)
						try
						{
							c.BeginInvoke(d,new object[]{(FeaturePointData)_FrameBuffer.Clone()});
						}
						catch {}
				}

			_FPS++;
		}

		private void Track()
		{
			//long tckStart = System.DateTime.Now.Ticks;
			if (!_TrackerInitialized)
			{
				//Check if every Frame exists before initializing tracker
				if (_FrameCur!=null && _FrameNew!=null && _FrameOld != null)
				{
					_Tracker.initialize(_FrameOld,_FrameCur,_FrameNew);

					_Tracker.postProcessFrame(_FrameOld);
					_TrackerInitialized = true;				
				}
			}
			else 
			{
				if (_FrameNew!=null)
				{
//					if (_DoubleFrameCheck)
//					{
//						//Only if the error between 2 pictures is larger than our threshold
//						//we will analyse the new frame with the tracker else it is considered 
//						//a doubleframe and we manually connect identical Featurepoints.
//						if (LeastSquareError()<1.0)
//						{
//							_Tracker.processDoubleFrame(_FrameCur,_FrameNew);
//
//						}
//						else
//						{
//							_Tracker.processFrame(_FrameOld,_FrameCur,_FrameNew);
//						}
//					}
//					else
//					{
					_Tracker.processFrame(_FrameOld,_FrameCur,_FrameNew);
//					}

					_Tracker.postProcessFrame(_FrameOld);

				}
			}

			//System.TimeSpan dT = new System.TimeSpan(System.DateTime.Now.Ticks - tckStart);
			//Console.WriteLine(dT.TotalMilliseconds);
			
		}
//		public void Finish()
//		{
//			_TrackerR.Finish();
//			_TrackerG.Finish();
//			_TrackerB.Finish();
//
//		}


		private double	LeastSquareError()
		{
			double Error = 0.0;
			Error += CalcLeastSquareError(_FrameCur.Pred,_FrameNew.Pred);
			Error += CalcLeastSquareError(_FrameCur.Pgreen,_FrameNew.Pgreen);
			Error += CalcLeastSquareError(_FrameCur.Pblue,_FrameNew.Pblue);

			return Error;
		}

		//Calculates LeastSquare of 2 FeaturePointSets
		private  double CalcLeastSquareError (SortedList a,  SortedList b)
		{
			SortedList S;
			SortedList s;

			int SCount;
			int sCount;

			double r = 0.0f;
			int dP;					//Difference in FeaturePoints between a and b
			float error=0.0f;
			
			if (a.Count > b.Count)
			{
				S= a;
				s= b;
			}
			else
			{
				s= a;
				S= b;
			}

			SCount = S.Count;
			sCount = s.Count;

			dP = SCount - sCount;

			FeaturePoint P;
			FeaturePoint p;

			

			for (int i = 0;i<sCount; i++)
			{
				p = (FeaturePoint)s.GetByIndex(i);

				error = 1000000.0f;
				for (int j=0; j<=dP; j++)
				{
					P = (FeaturePoint)S.GetByIndex(i+j);
					error = System.Math.Min(error,(P.X -  p.X) * (P.X -  p.X)+ (P.Y -  p.Y) * (P.Y -  p.Y));
				}
			
				r+= error;

			}

			return r;
		
		}

		/// <summary>
		/// FPS Refresher Function - announces current tracker working speed to all connected eventhandlers.
		/// </summary>
		private void FPSRefresher()
		{
			long Now;
			
			while(true)
			{
				Now = System.DateTime.Now.Ticks;

				if (RefreshFPS !=null) {RefreshFPS(_FPS);}
				_FPS = 0;
				Thread.Sleep(1000);
			}
		}

		#region IDisposable Member

		public void Dispose()
		{

			_thrFPS.Abort();
		}

		#endregion

	}
}
