using System;
using System.Net.Sockets;
using System.IO;
using System.Collections;
using System.Threading;

namespace MT.Filter.Server
{
/// <summary>
/// MTrackServer
/// This class provides a simple Serversystem. Setting a listener in the constructor
/// will automatically start a server. When a new connection is created, the server
/// will raise the NewConnection Event, announcing the ID(string) of the connection.
/// From that point in time on, a NewMessage Event is raised whenever a new Line character
/// is detected in the network stream.
/// The connection can be closed
/// </summary>
	public class Server: IDisposable
	{
		//declare Event Delegates
		public delegate void NewConnectionDelegate(string conID);
		public delegate void ClosedConnectionDelegate(string conID);
		public delegate void NewMessageDelegate(string conID, string message);

		//declare Events
		public event NewConnectionDelegate NewConnection ;
		public event ClosedConnectionDelegate ClosedConnection;
		public event NewMessageDelegate NewMessage;

		private Hashtable	_activeCons = new Hashtable();
		private Hashtable	_myThreads = new Hashtable();
		private TcpListener	_Listener;

		//Parameter for ReceiveLooper - should be reset to "" after being used
		private string	_conID="";

		public Server(int Port)
		{
			//TcpListener is listening on the given port...                                                 
			_Listener = new TcpListener(System.Net.IPAddress.Any, Port);
			_Listener.Start();    

			
			Thread thrListener = new Thread(new ThreadStart(Listener));
			thrListener.Name = "trhListener";
			_myThreads.Add(thrListener.Name, thrListener);
			thrListener.Start();
			Thread.Sleep(0);
		}

		//Contains ListenerLoop which checks for new connection requests.
		//Every time a new connection request arrives it checks its conID
		//and creates and starts a Thread for the ReceiverLoop.
		private void Listener()
		{

			TcpClient		incomingCon;
			NetworkStream	netStream;
			StreamReader	netRead;
			string			ConID;

			//Enter Message Loop
			while (true)
			{
				//AcceptTcpClient is a blocking Function which only continues,
				//when a new Connection Request arrives.
				incomingCon = _Listener.AcceptTcpClient();
				netStream = incomingCon.GetStream();
				netRead = new StreamReader(netStream);

				//The first line of a new connection must contain the connection ID
				ConID = netRead.ReadLine();

				//If ConID is already used close the new connection - else create new Receiver
				if (_activeCons.ContainsKey(ConID)){incomingCon.Close();}
				else
				{
					//Add new Connection to the Connection Pool
					_activeCons.Add(ConID,incomingCon);

					//Create new MessageLoop Thread for the Connection
					//1. Prepare Threadparameters - sleep while parameters are not reset -> synchronising
					while (_conID!=""){Thread.Sleep(0);} 
					_conID= ConID;

					//2. Announce new connection by raising event
					NewConnection(ConID);

					//3. Create Thread, name it and start it
					Thread thrReceiver = new Thread(new ThreadStart(Receiver));
					thrReceiver.Name = "thr" + ConID;
					_myThreads.Add(thrReceiver.Name, thrReceiver);

					thrReceiver.Start();
					Thread.Sleep(0);


				}
			}
		}

		//Contains ReceiverLoop which checks an active connection for new messages
		//this method is invoked for every new connection by a seperate thread
		//Since ThreadProcs cannot contain parameters, the parameters are stored
		//in public variables which are read out and then reset to allow new calls
		private void Receiver()
		{
			NetworkStream	netStream;
			StreamReader	netRead;
			TcpClient		currentCon;
			string			message;

			//Read Loop Parameters
			string			conID = _conID;

			//Reset Parameters
			_conID = "";

			try
			{
				//Prepare MessageLoop
				currentCon = (TcpClient)_activeCons[conID];
				netStream = currentCon.GetStream();
				netRead = new StreamReader(netStream);

			
				//Enter Messageloop
				do
				{
					//Read Message - ReadLine is a blocking Function
					//the thread is blocked as long as there is no 
					//new line character in the stream
					message = netRead.ReadLine();

					//Announce Message by raising event
					try
					{
						if (message!=null) 
							NewMessage(conID,message);
					}
					catch (Exception ex)
					{
					Console.WriteLine("Server.NewMessage - An exception occured: {0}",ex.Message);
					}
					

				} while (message!=null);
				//If message is null, the connection has been closed

			}
			catch{}

			//Remove Active Connection
			_activeCons.Remove(conID);
			_myThreads.Remove(Thread.CurrentThread.Name);

			//Announce Closed Connection
			ClosedConnection(conID);

		}

		#region IDisposable Members

		public void Dispose()
		{
			//Abort all Message Loops
			foreach(Thread thr in _myThreads.Values)
			{
				thr.Abort();
				Console.WriteLine("Aborted Thread " + thr.Name) ;      

			}

			_myThreads.Clear();
			_myThreads = null;

			//Close all Connections
			foreach(TcpClient tcl in _activeCons.Values)
			{	tcl.Close();
			}

			_activeCons.Clear();
			_activeCons = null;


			_Listener.Stop();
			_Listener = null;

		}

		#endregion
	}
}
