//  2001 IDesign Inc. All rights reserved 
//Questions? Comments? go to 
//http://www.idesign.net

using System;
using System.Diagnostics;
using System.Collections;
using System.ComponentModel;
using System.Threading;
using System.Security.Permissions;

namespace MT.Synchronization
{

	[Serializable]
	internal class WorkItem: IAsyncResult
	{
		object[] m_Args;
		object m_AsyncState;
		bool m_Completed;
		Delegate m_Method;
		ManualResetEvent m_Event;
		object m_MethodReturnedValue;
		internal WorkItem(object AsyncState,Delegate method,object[] args)
		{
			m_AsyncState = AsyncState;
			m_Method = method;
			m_Args =  args;
			m_Event = new ManualResetEvent(false);
			m_Completed = false;
		}
		//IAsyncResult properties 
		object IAsyncResult.AsyncState
		{ 
			get
			{
				return m_AsyncState;
			}
		}
		WaitHandle IAsyncResult.AsyncWaitHandle
		{ 
			get
			{
				return m_Event;
			}
		}
		bool IAsyncResult.CompletedSynchronously
		{
			get
			{
				return false;
			}
		}
		bool IAsyncResult.IsCompleted
		{ 
			get
			{
				return Completed;
			}
		}
		bool Completed
		{
			get
			{
				lock(this)
				{
					return m_Completed;
				}
			}
			set
			{
				lock(this)
				{
					m_Completed = value;
				}
			}
		}

		//This method is called on the worker thread to execute the method
		internal void CallBack()
		{
			MethodReturnedValue = m_Method.DynamicInvoke(m_Args);
			//Method is done. Signal the world
			m_Event.Set();
			Completed = true;
		}
		internal object MethodReturnedValue
		{
			get
			{
				object methodReturnedValue;
				lock(this)
				{
					methodReturnedValue = m_MethodReturnedValue;
				}
				return methodReturnedValue;
			}
			set
			{
				lock(this)
				{
					m_MethodReturnedValue = value;
				}
			}

		}
	}
	public class WorkerThread 
	{
		public Thread m_ThreadObj;
		bool m_EndLoop;
		Mutex m_EndLoopMutex;
		AutoResetEvent m_ItemAdded;
		object m_Synchronizer;
		Queue m_WorkItemQueue;

		public object Synchronizer
		{ 
			get {	return m_Synchronizer; }
		}

		internal void QueueWorkItem(WorkItem workItem)
		{
			lock(m_WorkItemQueue.SyncRoot)
			{
				m_WorkItemQueue.Enqueue(workItem);
				m_ItemAdded.Set();
			}
		}
		public WorkerThread(object synchronizer)
		{
			m_Synchronizer = synchronizer;
			m_EndLoop = false;
			m_ThreadObj = null;
			m_EndLoopMutex = new Mutex();
			m_ItemAdded = new AutoResetEvent(false);
			m_WorkItemQueue = new Queue();
			CreateThread(true);
		}
		bool EndLoop
		{
			set
			{
				m_EndLoopMutex.WaitOne();
				m_EndLoop = value;
				m_EndLoopMutex.ReleaseMutex();
			}
			get
			{
				bool result = false;
				m_EndLoopMutex.WaitOne();
				result = m_EndLoop;
				m_EndLoopMutex.ReleaseMutex();
				return result;
			}
		}
		Thread CreateThread(bool autoStart)
		{
			if(m_ThreadObj != null)
			{
				Debug.Assert(false);
				return m_ThreadObj;
			}
			ThreadStart threadStart = new ThreadStart(Run);
			m_ThreadObj = new Thread(threadStart);
			m_ThreadObj.Name = "Synchronizer Worker Thread";
			if(autoStart == true)
			{
				m_ThreadObj.Start();
			}
			return m_ThreadObj;
		}
		void Start()
		{
			Debug.Assert(m_ThreadObj != null);
			Debug.Assert(m_ThreadObj.IsAlive == false);
			m_ThreadObj.Start();
		}
		bool QueueEmpty
		{
			get
			{
				lock(m_WorkItemQueue.SyncRoot)
				{
					if(m_WorkItemQueue.Count > 0)
					{
						return false;
					}
					return true;
				}
			}
		}
		WorkItem GetNext()
		{
			if(QueueEmpty)
			{
				return null;
			}
			lock(m_WorkItemQueue.SyncRoot)
			{
				return (WorkItem)m_WorkItemQueue.Dequeue();
			}
		}
		void Run()
		{
			while(EndLoop == false)
			{ 
				while(QueueEmpty == false)
				{
					if(EndLoop == true)
					{
						return;
					}
					WorkItem workItem = GetNext();
					workItem.CallBack();
				}
				m_ItemAdded.WaitOne();
			}
		}
		public void Kill()
		{ 
			//Kill is called on client thread - must use cached thread object
			Debug.Assert(m_ThreadObj != null);
			if(m_ThreadObj.IsAlive == false)
			{
				return;
			}
			EndLoop = true;
			m_ItemAdded.Set();

			//Wait for thread to die
			m_ThreadObj.Join();
			if(m_EndLoopMutex != null)
			{
				m_EndLoopMutex.Close();
			}
			if(m_ItemAdded != null)
			{
				m_ItemAdded.Close();
			}
		}
	}

//	internal interface ISynchronizer: ISynchronizeInvoke, IDisposable
//	{
//		WorkerThread workerThread
//		{	get;
//		}
//	}

	[SecurityPermission(SecurityAction.Demand,ControlThread = true)]
	public class Synchronizer: ISynchronizeInvoke, IDisposable
	{
		protected WorkerThread m_WorkerThread;

		public bool InvokeRequired 
		{
			get
			{
				bool res = Thread.CurrentThread.GetHashCode() == m_WorkerThread.GetHashCode();
				return  ! res;
			}
		}
		public IAsyncResult BeginInvoke(Delegate method, object[] args)
		{
			WorkItem result = new WorkItem(null,method,args);
			m_WorkerThread.QueueWorkItem(result);
			return result;
		}
		public object EndInvoke(IAsyncResult result)
		{
			result.AsyncWaitHandle.WaitOne();
			WorkItem workItem = (WorkItem)result;
			return  workItem.MethodReturnedValue;
		}
		public object Invoke(Delegate method, object[] args)
		{
			IAsyncResult asyncResult;
			asyncResult = BeginInvoke(method,args);
			return EndInvoke(asyncResult);
		}
		public Synchronizer()
		{ 
			m_WorkerThread = new WorkerThread(this);
		}
		protected Synchronizer(WorkerThread workerThread)
		{
			m_WorkerThread = workerThread;
		}
		~Synchronizer()
		{
		}
		public void Dispose()
		{
			if (m_WorkerThread.Synchronizer == this)
				m_WorkerThread.Kill();
		}

	}
}