using System;
using System.IO;

namespace tsai
{
	[Serializable]
	public enum eTsaiOrientation
	{
		x, y, z
	}
	/// <summary>
	/// Zusammenfassung fr tsai_converter.
	/// </summary>
	[Serializable]
	public class tsai_converter
	{
		[Serializable]
		protected struct camera_parameters 
		{
			public double	Ncx;	//	[sel]		Number of sensor elements in camera's x direction
			public double	Nfx;	//	[pix]		Number of pixels in frame grabber's x direction
			public double	dx;		//	[mm/sel]	X dimension of camera's sensor element (in mm)
			public double	dy;		//	[mm/sel]	Y dimension of camera's sensor element (in mm)
			public double	dpx;	//	[mm/pix]	Effective X dimension of pixel in frame grabber
			public double	dpy;	//	[mm/pix]	Effective Y dimension of pixel in frame grabber
			public double	Cx;		//	[pix]		Z axis intercept of camera coordinate system
			public double	Cy;		//	[pix]		Z axis intercept of camera coordinate system
			public double	sx;		//	[]			Scale factor to compensate for any error in dpx
		};

		[Serializable]
		public struct calibration_constants
		{
			public double	f;		//	[mm]
			public double	kappa1;	//	[1/mm^2]
			public double	p1;		//	[1/mm]
			public double	p2;		//	[1/mm]
			public double	Tx;		//	[mm]
			public double	Ty;		//	[mm]
			public double	Tz;		//	[mm]
			public double	Rx;		//	[rad]
			public double	Ry;		//	[rad]
			public double	Rz;		//	[rad]
			public double	r1;		//	[]
			public double	r2;		//	[]
			public double	r3;		//	[]
			public double	r4;		//	[]
			public double	r5;		//	[]
			public double	r6;		//	[]
			public double	r7;		//	[]
			public double	r8;		//	[]
			public double	r9;		//	[]
		};

		protected eTsaiOrientation _orientation;
		protected camera_parameters _cp;
		public calibration_constants _cc;

		public void get_camera_position(ref double xc, ref double yc, ref double zc)
		{
			xc = _cc.r1*(-_cc.Tx) + _cc.r4*(-_cc.Ty) + _cc.r7*(-_cc.Tz);
			yc = _cc.r2*(-_cc.Tx) + _cc.r5*(-_cc.Ty) + _cc.r8*(-_cc.Tz);
			zc = _cc.r3*(-_cc.Tx) + _cc.r6*(-_cc.Ty) + _cc.r9*(-_cc.Tz);
		}


		private void distorted_to_undistorted_sensor_coord(double Xd, double Yd, ref double Xu, ref double Yu)
		{
			double distortion_factor;
			distortion_factor = 1 + _cc.kappa1 * (Xd*Xd + Yd*Yd);
			Xu = Xd * distortion_factor;
			Yu = Yd * distortion_factor;
		}


		private void undistorted_to_distorted_sensor_coord(double Xu, double Yu, ref double Xd, ref double Yd)
		{
			const double SQRT3 = 1.732050807568877293527446341505872366943;
			
			double	Ru,
					Rd,
					lambda,
					c,
					d,
					Q,
					R,
					D,
					S,
					T,
					sinT,
					cosT;
			
			if(((Xu == 0) && (Yu == 0)) || (_cc.kappa1 == 0))
			{
				Xd = Xu;
				Yd = Yu;
				return;
			}

			Ru = Math.Sqrt(Xu*Xu+Yu*Yu);
			
			c = 1 / _cc.kappa1;
			d = -c * Ru;

			Q = c / 3;
			R = -d / 2;
			D = (Q * Q * Q) + (R * R);
			
			if(D >= 0)
			{
				D	= Math.Sqrt(D);
				S	= Math.Exp(Math.Log(R + D) / 3); // exp(ln(x) / n) n = n-th root of x
				if( R-D < 0 )
					T=-1*Math.Exp(Math.Log(-1*(R - D)) / 3);
				else
					T	= Math.Exp(Math.Log(R - D) / 3); // exp(ln(x) / n)
				Rd	= S + T;

				if(Rd < 0)
				{
					Rd = Math.Sqrt(-1 / (3 * _cc.kappa1));
					Console.WriteLine("\nWarning: undistorted image point to distorted image point mapping");
					Console.WriteLine("           maximum barrel distortion radius of %lf\n", Rd);
					Console.WriteLine("           (Xu = %lf, Yu = %lf) -> (Xd = %lf, Yd = %lf)\n\n",
						Xu, Yu, Xu * Rd / Ru, Yu * Rd / Ru);
				}
			}	
			else
			{
				D = Math.Sqrt(-D);
				S = Math.Pow(10,((Math.Log10(Math.Sqrt((R*R) + (D*D))))/3));
				T = Math.Atan2(D,R) / 3;
				sinT = Math.Sin(T);
				cosT = Math.Cos(T);

				Rd = -S * cosT + SQRT3 * S * sinT;
			}

			lambda = Rd / Ru;
			
			Xd = Xu * lambda;
			Yd = Yu * lambda;
		}


		public void world_coord_to_image_coord(double xw, double yw, double zw, ref double Xf, ref double Yf)
		{
			double		xc,
						yc,
						zc,
						Xu,
						Yu;
			double		Xd = new double();
			double		Yd = new double();

			xc = _cc.r1 * xw + _cc.r2 * yw + _cc.r3 * zw + _cc.Tx;
			yc = _cc.r4 * xw + _cc.r5 * yw + _cc.r6 * zw + _cc.Ty;
			zc = _cc.r7 * xw + _cc.r8 * yw + _cc.r9 * zw + _cc.Tz;

			Xu = _cc.f * xc / zc;
			Yu = _cc.f * yc / zc;

			undistorted_to_distorted_sensor_coord(Xu, Yu, ref Xd, ref Yd);

			Xf = Xd * _cp.sx / _cp.dpx + _cp.Cx;
			Yf = Yd / _cp.dpy + _cp.Cy;
		}


		public void image_coord_to_world_coord(double Xfd, double Yfd, double zw, ref double xw, ref double yw)
		{
			double	Xd,
					Yd,
					common_denominator;
			double	Xu = new double();
			double	Yu = new double();
					

			Xd = _cp.dpx * (Xfd - _cp.Cx) / _cp.sx;
			Yd = _cp.dpy * (Yfd - _cp.Cy);

			distorted_to_undistorted_sensor_coord(Xd, Yd, ref Xu, ref Yu);

			common_denominator = ((_cc.r1 * _cc.r8 - _cc.r2 * _cc.r7) * Yu +
								  (_cc.r5 * _cc.r7 - _cc.r4 * _cc.r8) * Xu -
								  _cc.f * _cc.r1 * _cc.r5 + _cc.f * _cc.r2 * _cc.r4);


			double ax = ((_cc.r2 * _cc.r9 - _cc.r3 * _cc.r8) * Yu +
						(_cc.r6 * _cc.r8 - _cc.r5 * _cc.r9) * Xu -
						_cc.f * _cc.r2 * _cc.r6 + _cc.f * _cc.r3 * _cc.r5);

			double bx = (_cc.r2 * _cc.Tz - _cc.r8 * _cc.Tx) * Yu +
						(_cc.r8 * _cc.Ty - _cc.r5 * _cc.Tz) * Xu -
						_cc.f * _cc.r2 * _cc.Ty + _cc.f * _cc.r5 * _cc.Tx;

			double ay = ((_cc.r1 * _cc.r9 - _cc.r3 * _cc.r7) * Yu +
						(_cc.r6 * _cc.r7 - _cc.r4 * _cc.r9) * Xu -
						_cc.f * _cc.r1 * _cc.r6 + _cc.f * _cc.r3 * _cc.r4);

			double by = (_cc.r1 * _cc.Tz - _cc.r7 * _cc.Tx) * Yu +
						(_cc.r7 * _cc.Ty - _cc.r4 * _cc.Tz) * Xu -
						_cc.f * _cc.r1 * _cc.Ty + _cc.f * _cc.r4 * _cc.Tx;

			double c = common_denominator;


			xw = ( ax*zw + bx ) / c ;
			yw = -( ay*zw + by ) / c ;
/*
			xw = (((_cc.r2 * _cc.r9 - _cc.r3 * _cc.r8) * Yu +
				(_cc.r6 * _cc.r8 - _cc.r5 * _cc.r9) * Xu -
				_cc.f * _cc.r2 * _cc.r6 + _cc.f * _cc.r3 * _cc.r5) * zw +
				(_cc.r2 * _cc.Tz - _cc.r8 * _cc.Tx) * Yu +
				(_cc.r8 * _cc.Ty - _cc.r5 * _cc.Tz) * Xu -
				_cc.f * _cc.r2 * _cc.Ty + _cc.f * _cc.r5 * _cc.Tx) / common_denominator;

			yw = -(((_cc.r1 * _cc.r9 - _cc.r3 * _cc.r7) * Yu +
				(_cc.r6 * _cc.r7 - _cc.r4 * _cc.r9) * Xu -
				_cc.f * _cc.r1 * _cc.r6 + _cc.f * _cc.r3 * _cc.r4) * zw +
				(_cc.r1 * _cc.Tz - _cc.r7 * _cc.Tx) * Yu +
				(_cc.r7 * _cc.Ty - _cc.r4 * _cc.Tz) * Xu -
				_cc.f * _cc.r1 * _cc.Ty + _cc.f * _cc.r4 * _cc.Tx) / common_denominator;
*/					
		}


		public void image_coord_to_world_coord(double Xfd, double Yfd, ref double zw, double xw, ref double yw)
		{
			double	Xd,
					Yd,
					common_denominator;
			double	Xu = new double();
			double	Yu = new double();
					

			Xd = _cp.dpx * (Xfd - _cp.Cx) / _cp.sx;
			Yd = _cp.dpy * (Yfd - _cp.Cy);

			distorted_to_undistorted_sensor_coord(Xd, Yd, ref Xu, ref Yu);

			common_denominator = ((_cc.r1 * _cc.r8 - _cc.r2 * _cc.r7) * Yu +
				(_cc.r5 * _cc.r7 - _cc.r4 * _cc.r8) * Xu -
				_cc.f * _cc.r1 * _cc.r5 + _cc.f * _cc.r2 * _cc.r4);

			double ax = ((_cc.r2 * _cc.r9 - _cc.r3 * _cc.r8) * Yu +
						(_cc.r6 * _cc.r8 - _cc.r5 * _cc.r9) * Xu -
						_cc.f * _cc.r2 * _cc.r6 + _cc.f * _cc.r3 * _cc.r5);

			double bx = (_cc.r2 * _cc.Tz - _cc.r8 * _cc.Tx) * Yu +
						(_cc.r8 * _cc.Ty - _cc.r5 * _cc.Tz) * Xu -
						_cc.f * _cc.r2 * _cc.Ty + _cc.f * _cc.r5 * _cc.Tx;

			double ay = ((_cc.r1 * _cc.r9 - _cc.r3 * _cc.r7) * Yu +
						(_cc.r6 * _cc.r7 - _cc.r4 * _cc.r9) * Xu -
						_cc.f * _cc.r1 * _cc.r6 + _cc.f * _cc.r3 * _cc.r4);

			double by = (_cc.r1 * _cc.Tz - _cc.r7 * _cc.Tx) * Yu +
						(_cc.r7 * _cc.Ty - _cc.r4 * _cc.Tz) * Xu -
						_cc.f * _cc.r1 * _cc.Ty + _cc.f * _cc.r4 * _cc.Tx;

			double c = common_denominator;

			zw =  (  c*xw - bx ) / ax;

			yw = -( ay*zw + by ) / c ;
					
		}


		public void image_coord_to_world_coord(double Xfd, double Yfd, ref double zw, ref double xw, double yw)
		{
			double	Xd,
					Yd,
					common_denominator;
			double	Xu = new double();
			double	Yu = new double();
					

			Xd = _cp.dpx * (Xfd - _cp.Cx) / _cp.sx;
			Yd = _cp.dpy * (Yfd - _cp.Cy);

			distorted_to_undistorted_sensor_coord(Xd, Yd, ref Xu, ref Yu);

			common_denominator = ((_cc.r1 * _cc.r8 - _cc.r2 * _cc.r7) * Yu +
				(_cc.r5 * _cc.r7 - _cc.r4 * _cc.r8) * Xu -
				_cc.f * _cc.r1 * _cc.r5 + _cc.f * _cc.r2 * _cc.r4);

			double ax = ((_cc.r2 * _cc.r9 - _cc.r3 * _cc.r8) * Yu +
				(_cc.r6 * _cc.r8 - _cc.r5 * _cc.r9) * Xu -
				_cc.f * _cc.r2 * _cc.r6 + _cc.f * _cc.r3 * _cc.r5);

			double bx = (_cc.r2 * _cc.Tz - _cc.r8 * _cc.Tx) * Yu +
				(_cc.r8 * _cc.Ty - _cc.r5 * _cc.Tz) * Xu -
				_cc.f * _cc.r2 * _cc.Ty + _cc.f * _cc.r5 * _cc.Tx;

			double ay = ((_cc.r1 * _cc.r9 - _cc.r3 * _cc.r7) * Yu +
				(_cc.r6 * _cc.r7 - _cc.r4 * _cc.r9) * Xu -
				_cc.f * _cc.r1 * _cc.r6 + _cc.f * _cc.r3 * _cc.r4);

			double by = (_cc.r1 * _cc.Tz - _cc.r7 * _cc.Tx) * Yu +
				(_cc.r7 * _cc.Ty - _cc.r4 * _cc.Tz) * Xu -
				_cc.f * _cc.r1 * _cc.Ty + _cc.f * _cc.r4 * _cc.Tx;

			double c = common_denominator;

			zw = -(  c*yw + by ) / ay;

			xw =  ( ax*zw + bx ) / c ;
					
		}


		public void image_coord_to_world_coord(double Xfd, double Yfd, ref double zw, ref double xw, ref double yw)
		{
			switch (_orientation)
			{
				case eTsaiOrientation.x: 
					xw = 10000.0;
					image_coord_to_world_coord(Xfd, Yfd, ref zw, xw, ref yw);
					break;
				case eTsaiOrientation.y:
					yw = 10000.0;
					image_coord_to_world_coord(Xfd, Yfd, ref zw, ref xw, yw);
					break;
				case eTsaiOrientation.z: 
					zw = 10000.0;
					image_coord_to_world_coord(Xfd, Yfd, zw, ref xw, ref yw);
					break;
			}
//			yw = 10000.0;
//			image_coord_to_world_coord(Xfd, Yfd, ref zw, ref xw, yw);
		}

		/// <summary>
		/// Standardkonstruktor tsai_converter(string filename)
		/// </summary>
		public tsai_converter(string filename, eTsaiOrientation orientation)
		{
			_orientation = orientation;
			TextReader tr = new StreamReader(filename);
			
			double	sa,
					ca,
					sb,
					cb,
					sg,
					cg;

			_cp.Ncx		= double.Parse(tr.ReadLine());
			_cp.Nfx		= double.Parse(tr.ReadLine());
			_cp.dx		= double.Parse(tr.ReadLine());
			_cp.dy		= double.Parse(tr.ReadLine());
			_cp.dpx		= double.Parse(tr.ReadLine());
			_cp.dpy		= double.Parse(tr.ReadLine());
			_cp.Cx		= double.Parse(tr.ReadLine());
			_cp.Cy		= double.Parse(tr.ReadLine());
			_cp.sx		= double.Parse(tr.ReadLine());

			_cc.f		= double.Parse(tr.ReadLine());
			_cc.kappa1	= double.Parse(tr.ReadLine());
			_cc.Tx		= double.Parse(tr.ReadLine());
			_cc.Ty		= double.Parse(tr.ReadLine());
			_cc.Tz		= double.Parse(tr.ReadLine());
			_cc.Rx		= double.Parse(tr.ReadLine());
			_cc.Ry		= double.Parse(tr.ReadLine());
			_cc.Rz		= double.Parse(tr.ReadLine());
			
			sa = Math.Sin(_cc.Rx);
			ca = Math.Cos(_cc.Rx);
			sb = Math.Sin(_cc.Ry);
			cb = Math.Cos(_cc.Ry);
			sg = Math.Sin(_cc.Rz);
			cg = Math.Cos(_cc.Rz);

			_cc.r1 = cb * cg;
			_cc.r2 = cg * sa * sb - ca * sg;
			_cc.r3 = sa * sg + ca * cg * sb;
			_cc.r4 = cb * sg;
			_cc.r5 = sa * sb * sg + ca * cg;
			_cc.r6 = ca * sb * sg - cg * sa;
			_cc.r7 = -sb;
			_cc.r8 = cb * sa;
			_cc.r9 = ca * cb;

			_cc.p1 = double.Parse(tr.ReadLine());
			_cc.p2 = double.Parse(tr.ReadLine());

			tr.Close();
		}

		public tsai_converter(tsai_converter tsai)
		{
			if (tsai!=null)
			{
				_cc			 = tsai._cc;
				_cp			 = tsai._cp;
				_orientation = tsai._orientation;
			}
		}

		#region "Properties"
		public calibration_constants CC
		{
			get{return _cc;}
		}
		#endregion
	}
}